import json
from pathlib import Path

from pyfakefs.fake_filesystem import FakeFilesystem

from ggshield.core.scan.id_cache import IDCache
from tests.unit.conftest import make_fake_path_inaccessible


def test_add(fs: FakeFilesystem):
    """
    GIVEN a cache instance
    WHEN add() is called
    THEN a cache file is created
    """
    cache_path = Path("/some_dir/cache.json")
    cache = IDCache(cache_path)

    some_id = "iam_an_id"
    cache.add(some_id)

    assert cache_path.exists()
    ids = json.loads(cache_path.read_text())
    assert ids == [some_id]


def test_contains(fs: FakeFilesystem):
    """
    GIVEN a cache file
    WHEN __contains__() is called
    THEN the response is what was expected
    """
    cache_path = Path("/some_dir/cache.json")
    some_id = "iam_an_id"

    # Create the file on the disk
    IDCache(cache_path).add(some_id)
    assert cache_path.exists()

    cache = IDCache(cache_path)
    assert some_id in cache
    assert "unknown_id" not in cache


def test_does_not_fail_if_cache_file_cannot_be_written(caplog, fs: FakeFilesystem):
    """
    GIVEN an instance of IDCache created for a path which cannot be created
    WHEN adding an entry
    THEN it does not raise an exception
    AND a log message is written
    """
    cache_path = Path("/some_dir/cache.json")
    cache_path.parent.mkdir()
    make_fake_path_inaccessible(fs, cache_path.parent)

    cache = IDCache(cache_path)
    cache.add("clean_id")
    log_record = caplog.records[0]
    assert log_record.levelname == "WARNING"
    assert "Failed to save" in log_record.message
    assert len(caplog.records) == 1


def test_does_not_fail_if_cache_file_cannot_be_read(caplog, fs: FakeFilesystem):
    """
    GIVEN a non-readable cache file
    WHEN an instance of IDCache is created using it
    THEN it does not crash
    AND a log message is written
    AND the cache can be queried
    """
    cache_path = Path("/some_dir/cache.json")
    cache_path.parent.mkdir()
    cache_path.write_text('["hello"]')
    make_fake_path_inaccessible(fs, cache_path)

    cache = IDCache(cache_path)
    log_record = caplog.records[0]
    assert log_record.levelname == "WARNING"
    assert "Failed to load" in log_record.message
    assert len(caplog.records) == 1

    assert "hello" not in cache



import logging
from pathlib import Path
from subprocess import CalledProcessError

import pytest

from tests.conftest import GG_VALID_TOKEN
from tests.functional.utils import recreate_censored_content
from tests.repository import Repository


HOOK_CONTENT = """#!/bin/sh
set -e
ggshield secret scan pre-receive
"""


def test_scan_prereceive(tmp_path: Path) -> None:
    # GIVEN a remote repository
    remote_repo = Repository.create(tmp_path / "remote", bare=True)

    # AND a local clone
    local_repo = Repository.clone(remote_repo.path, tmp_path / "local")

    # AND ggshield installed as a pre-receive hook
    hook_path = remote_repo.path / "hooks" / "pre-receive"
    hook_path.write_text(HOOK_CONTENT)
    hook_path.chmod(0o755)

    # AND a secret committed
    secret_file = local_repo.path / "secret.conf"
    secret_content = f"password = {GG_VALID_TOKEN}"
    secret_file.write_text(secret_content)
    local_repo.add("secret.conf")
    local_repo.create_commit()

    # WHEN I try to push
    # THEN the hook prevents the push
    with pytest.raises(CalledProcessError) as exc:
        local_repo.push()

    # AND the error message contains the leaked secret
    stderr = exc.value.stderr.decode()
    assert recreate_censored_content(secret_content, GG_VALID_TOKEN) in stderr


def test_scan_prereceive_branch_without_new_commits(tmp_path: Path) -> None:
    # GIVEN a remote repository
    remote_repo = Repository.create(tmp_path / "remote", bare=True)

    # AND a local clone
    local_repo = Repository.clone(remote_repo.path, tmp_path / "local")

    # Add a commit to the remote repository, otherwise git complains the branch does not
    # contain anything
    local_repo.create_commit()
    local_repo.push()

    # AND ggshield installed as a pre-receive hook
    hook_path = remote_repo.path / "hooks" / "pre-receive"
    hook_path.write_text(HOOK_CONTENT)
    hook_path.chmod(0o755)

    # AND a branch without new commits
    branch_name = "topic"
    local_repo.create_branch(branch_name)

    # WHEN I try to push the branch
    # THEN the hook does not crash
    local_repo.push("-u", "origin", branch_name)


def test_scan_prereceive_push_force(tmp_path: Path) -> None:
    # GIVEN a remote repository
    remote_repo = Repository.create(tmp_path / "remote", bare=True)

    # AND a local clone
    local_repo = Repository.clone(remote_repo.path, tmp_path / "local")
    initial_sha = local_repo.create_commit("Initial commit")

    # AND a secret committed and pushed
    secret_file = local_repo.path / "secret.conf"
    secret_content = f"password = {GG_VALID_TOKEN}"
    secret_file.write_text(secret_content)
    local_repo.add("secret.conf")
    local_repo.create_commit()
    local_repo.push()

    # AND ggshield installed as a pre-receive hook
    hook_path = remote_repo.path / "hooks" / "pre-receive"
    hook_path.write_text(HOOK_CONTENT)
    hook_path.chmod(0o755)

    # AND a commit overwriting the leak commit
    local_repo.git("reset", "--hard", initial_sha)
    secret_file.write_text("password = $FROM_ENV")
    local_repo.add("secret.conf")
    local_repo.create_commit()

    # WHEN I push force to overwrite the commit with the secret
    # THEN the push is accepted because the commit containing the secret has not been
    # scanned
    local_repo.push("--force")


def test_scan_prereceive_timeout(
    tmp_path: Path, monkeypatch, slow_gitguardian_api: str, caplog
) -> None:
    # GIVEN a remote repository
    remote_repo = Repository.create(tmp_path / "remote", bare=True)

    # AND a local clone
    local_repo = Repository.clone(remote_repo.path, tmp_path / "local")

    # AND ggshield installed as a pre-receive hook
    hook_path = remote_repo.path / "hooks" / "pre-receive"
    hook_path.write_text(HOOK_CONTENT)
    hook_path.chmod(0o755)

    # AND a secret committed
    secret_file = local_repo.path / "secret.conf"
    secret_content = f"password = {GG_VALID_TOKEN}"
    secret_file.write_text(secret_content)
    local_repo.add("secret.conf")
    local_repo.create_commit()

    # WHEN I try to push
    # THEN the hook timeouts and allows the push
    with caplog.at_level(logging.WARNING):
        monkeypatch.setenv("GITGUARDIAN_API_URL", slow_gitguardian_api)
        monkeypatch.delenv("GITGUARDIAN_INSTANCE", raising=False)
        local_repo.push()

    # AND the error message contains timeout message
    assert any(
        "Pre-receive hook took too long" in record.message
        for record in caplog.records
        if record.levelname == "WARNING"
    )



from pathlib import Path
from typing import Any

import click


class RealPath(click.Path):
    """
    A click.Path which uses real Path objects
    """

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, path_type=Path, **kwargs)



import os
from pathlib import Path
from typing import Union
from unittest import mock

import pytest

from ggshield.core.scan.scan_context import ScanContext
from ggshield.core.scan.scan_mode import ScanMode
from tests.repository import Repository


DOMAIN = "https://github.com"
SLUG = "from-env/repository"
ENV_REPO_URL = f"{DOMAIN}/{SLUG}"
EXPECTED_HEADER_ENV = "github.com/from-env/repository"

REMOTE_URL = "https://github.com/from-remote/repository"
FULL_REMOTE_URL = "https://user:password@github.com:84/from-remote/repository.git"
EXPECTED_HEADER_REMOTE = "github.com/from-remote/repository"


@pytest.fixture(scope="module")
def fake_url_repo(tmp_path_factory: pytest.TempPathFactory) -> Repository:
    repo = Repository.create(tmp_path_factory.mktemp("fake_url_repo"))
    repo.git("remote", "add", "origin", REMOTE_URL)
    return repo


def _assert_repo_url_in_headers(context: ScanContext, expected_url: Union[Path, str]):
    assert context.get_http_headers().get("GGShield-Repository-URL") == str(
        expected_url
    )


def _assert_no_repo_url_in_headers(context: ScanContext):
    assert context.get_http_headers().get("GGShield-Repository-URL") is None


def test_scan_context_no_repo(
    tmp_path: Path,
):
    """
    GIVEN a directory which is not a git repo
    WHEN passing the local path to the scan context
    THEN there is no GGShield-Repository-URL in the headers
    """
    context = ScanContext(
        scan_mode=ScanMode.PATH,
        command_path="ggshield secret scan path",
        target_path=tmp_path,
    )
    _assert_no_repo_url_in_headers(context)


def test_scan_context_repository_url_parsed(fake_url_repo: Repository):
    """
    GIVEN a repository with a remote (url)
    WHEN passing the local path to the scan context
    THEN the remote url is found and simplified
    """
    context = ScanContext(
        scan_mode=ScanMode.PATH,
        command_path="ggshield secret scan path",
        target_path=fake_url_repo.path,
    )
    _assert_repo_url_in_headers(context, EXPECTED_HEADER_REMOTE)


@pytest.mark.parametrize(
    "env",
    (
        {"BUILD_BUILDID": "1", "BUILD_REPOSITORY_URI": ENV_REPO_URL},
        {"DRONE": "1", "DRONE_REPO_LINK": ENV_REPO_URL},
        {
            "GITHUB_ACTIONS": "1",
            "GITHUB_SERVER_URL": DOMAIN,
            "GITHUB_REPOSITORY": SLUG,
        },
        {"GITLAB_CI": "1", "CI_REPOSITORY_URL": ENV_REPO_URL},
        {"CIRCLECI": "1", "CIRCLE_REPOSITORY_URL": ENV_REPO_URL},
        {"BITBUCKET_COMMIT": "1", "BITBUCKET_GIT_HTTP_ORIGIN": ENV_REPO_URL},
    ),
)
def test_ci_repo_found(env, fake_url_repo: Repository) -> None:
    """
    GIVEN a repository with a remote url
    WHEN a repository url is found in the CI environment
    THEN it is sent instead of the remote url
    """
    with mock.patch.dict(os.environ, env, clear=True):
        context = ScanContext(
            scan_mode=ScanMode.CI,
            command_path="ggshield secret scan path",
            target_path=fake_url_repo.path,
        )
        _assert_repo_url_in_headers(context, EXPECTED_HEADER_ENV)


@pytest.mark.parametrize(
    "env",
    (
        {"BUILD_BUILDID": "1"},
        {"DRONE": "1"},
        {"GITHUB_ACTIONS": "1"},
        {"GITLAB_CI": "1"},
        {"CIRCLECI": "1"},
        {"BITBUCKET_COMMIT": "1"},
    ),
)
def test_ci_no_env(env, fake_url_repo: Repository) -> None:
    """
    GIVEN a repository with a remote url
    WHEN there is no repository url in the environment
    THEN the remote url is sent by default
    """
    # Copying the path is needed for windows to find git
    environ = {"PATH": os.environ.get("PATH"), **env}
    with mock.patch.dict(os.environ, environ, clear=True):
        context = ScanContext(
            scan_mode=ScanMode.PATH,
            command_path="ggshield secret scan path",
            target_path=fake_url_repo.path,
        )
        _assert_repo_url_in_headers(context, EXPECTED_HEADER_REMOTE)



from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, cast

import marshmallow_dataclass
from pygitguardian.models_utils import FromDictMixin, ToDictMixin

from ggshield.core.config.utils import (
    get_auth_config_filepath,
    load_yaml_dict,
    replace_dash_in_keys,
    save_yaml_dict,
)
from ggshield.core.errors import (
    AuthExpiredError,
    MissingTokenError,
    UnknownInstanceError,
)
from ggshield.utils.datetime import datetime_from_isoformat


@dataclass
class AccountConfig:
    workspace_id: int
    token: str
    type: str
    token_name: str
    expire_at: Optional[datetime]


@dataclass
class InstanceConfig:
    # Only handle 1 account per instance for the time being
    account: Optional[AccountConfig]
    url: str
    name: Optional[str] = None
    # The token lifetime. If it's set it overrides AuthConfig.default_token_lifetime
    default_token_lifetime: Optional[int] = None

    @property
    def expired(self) -> bool:
        return (
            self.account is not None
            and self.account.expire_at is not None
            and self.account.expire_at <= datetime.now(timezone.utc)
        )

    def init_account(self, token: str, token_data: Dict[str, Any]) -> None:
        """Initialize our account based on the token and the data received from the
        `token` endpoint"""
        expire_at_str = token_data.get("expire_at")
        expire_at = (
            None if expire_at_str is None else datetime_from_isoformat(expire_at_str)
        )
        self.account = AccountConfig(
            workspace_id=cast(int, token_data["account_id"]),
            token=token,
            expire_at=expire_at,
            token_name=token_data["name"],
            type=token_data["type"],
        )


def prepare_auth_config_dict_for_parse(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Since we support only one account for now, turn instances[].accounts keys into a
    single instances[].account key.

    Stop if there is not exactly one account.

    We replace `-` with `_` for compatibility reasons.
    """
    replace_dash_in_keys(data)
    data = deepcopy(data)
    try:
        instances = data["instances"]
    except KeyError:
        return data

    for instance in instances:
        accounts = instance.pop("accounts")
        assert (
            len(accounts) == 1
        ), "Each GitGuardian instance should have exactly one account"
        instance["account"] = accounts[0]

    return data


def prepare_auth_config_dict_for_save(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Does the opposite of `prepare_auth_config_dict_for_parse`: turn the
    instances[].account key into instances[].accounts keys.
    """
    data = deepcopy(data)
    try:
        instances = data["instances"]
    except KeyError:
        return data

    for instance in instances:
        account = instance.pop("account")
        instance["accounts"] = [account]

    return data


@dataclass
class AuthConfig(FromDictMixin, ToDictMixin):
    """
    Holds all declared GitGuardian instances and their tokens.
    Knows how to load and save them from the YAML file at get_auth_config_filepath().
    """

    default_token_lifetime: Optional[int] = None
    instances: List[InstanceConfig] = field(default_factory=list)

    @classmethod
    def load(cls) -> "AuthConfig":
        """Load the auth config from the app config file"""
        config_path = get_auth_config_filepath()

        data = load_yaml_dict(config_path)
        if data:
            data = prepare_auth_config_dict_for_parse(data)
            return cls.from_dict(data)
        return cls()

    def save(self) -> None:
        config_path = get_auth_config_filepath()
        data = prepare_auth_config_dict_for_save(self.to_dict())
        save_yaml_dict(data, config_path, restricted=True)

    def get_instance(self, instance_name: str) -> InstanceConfig:
        for instance in self.instances:
            if instance.name == instance_name or instance.url == instance_name:
                return instance
        else:
            raise UnknownInstanceError(instance=instance_name)

    def get_or_create_instance(self, instance_name: str) -> InstanceConfig:
        try:
            instance_config = self.get_instance(instance_name=instance_name)
        except UnknownInstanceError:
            # account is initialized as None because the instance must exist in
            # the config before using the client
            instance_config = InstanceConfig(account=None, url=instance_name)
            self.instances.append(instance_config)
        return instance_config

    def set_instance(self, instance_config: InstanceConfig) -> None:
        instance_name = instance_config.url
        for i, instance in enumerate(self.instances):
            if instance.url == instance_name or instance.name == instance_name:
                self.instances[i] = instance_config
                break
        else:
            self.instances.append(instance_config)

    def get_instance_token(self, instance_name: str) -> str:
        """
        Return the API token associated with the given instance if it is still valid.

        Raise AuthExpiredError if it is not.
        """
        instance = self.get_instance(instance_name)
        if instance.expired:
            raise AuthExpiredError(instance=instance_name)
        if instance.account is None:
            raise MissingTokenError(instance=instance_name)
        return instance.account.token


AuthConfig.SCHEMA = marshmallow_dataclass.class_schema(AuthConfig)()



import logging
import os
import platform
import re
import subprocess
from enum import Enum
from functools import lru_cache
from pathlib import Path
from shutil import which
from typing import List, Optional, Union

from ggshield.utils.os import getenv_int


EMPTY_SHA = "0000000000000000000000000000000000000000"
EMPTY_TREE = "4b825dc642cb6eb9a060e54bf8d69288fbee4904"

COMMAND_TIMEOUT = getenv_int("GG_GIT_TIMEOUT", 45)

logger = logging.getLogger(__name__)


class GitError(Exception):
    pass


class InvalidGitRefError(GitError):
    """
    Raised when the git reference does not exist
    """

    def __init__(self, ref: str):
        super().__init__(f"Not a git reference: {ref}.")


class NotAGitDirectory(GitError):
    def __init__(self):
        super().__init__("Not a git directory.")


class GitExecutableNotFound(GitError):
    pass


class GitCommandTimeoutExpired(GitError):
    pass


class Filemode(Enum):
    """
    Enum class for git filemode.
    """

    MODIFY = "modified file"
    DELETE = "deleted file"
    NEW = "new file"
    RENAME = "renamed file"
    FILE = "file"
    UNKNOWN = "unknown"


@lru_cache(None)
def _get_git_path() -> str:
    git_path = which("git")

    if git_path is None:
        raise GitExecutableNotFound("unable to find git executable in PATH/PATHEXT")

    # lower()ing these would provide additional coverage on case-
    # insensitive filesystems but detection is problematic
    git_path = os.path.abspath(git_path)
    cwd = os.getcwd()
    path_env = [
        os.path.abspath(p) for p in os.environ.get("PATH", "").split(os.pathsep)
    ]

    # git was found - ignore git in cwd if cwd not in PATH
    if cwd == os.path.dirname(git_path) and cwd not in path_env:
        raise GitExecutableNotFound("rejecting git executable in CWD not in PATH")

    logger.debug("Found git at %s", git_path)
    return git_path


@lru_cache(None)
def _git_rev_parse_absolute(option: str, wd_absolute: Path) -> Optional[str]:
    """
    Helper function for `_git_rev_parse` to only cache on absolute paths.
    """
    try:
        return git(["rev-parse", option], cwd=wd_absolute, log_stderr=False)
    except subprocess.CalledProcessError:
        return None


def _git_rev_parse(option: str, wd: Path) -> Optional[str]:
    return _git_rev_parse_absolute(option=option, wd_absolute=wd.resolve())


def get_new_branch_ci_commits(
    branch: str, wd: Path, remote: str = "origin"
) -> List[str]:
    """
    Returns a list of commits that only exist on the given branch.
    This is intended to be used for new branches only, in a CI env.
    """
    # https://stackoverflow.com/q/14848274
    refs_format = f"refs/remotes/{remote}/"
    all_branches = git(
        [
            "for-each-ref",
            "--format=%(refname)",
            refs_format,
        ],
        cwd=wd,
    ).splitlines()
    other_branches = (b for b in all_branches if b != f"{refs_format}{branch}")

    return git(
        ["log", "HEAD", "--not", *other_branches, "--format=format:%H"], cwd=wd
    ).splitlines()


def simplify_git_url(url: str) -> str:
    """
    Removes elements from the git remote url.
    - scheme
    - credentials
    - port
    - extension
    https://user:pass@mygitlab.corp.com:84/path/to/repo.git -> mygitlab.corp.com/toto/titi/tata
    """
    for pattern, replace in (
        (r"https?://", ""),  # Scheme
        (r".+@", ""),  # Credentials
        (r":\d*/", "/"),  # Port
        (r"\.git$", ""),  # Github/Gitlab/BitBucket extension (**.git)
        (r"/_git/", "/"),  # Azure Devops extension (**/_git/**)
        (":", "/"),  # Normalize ssh url to https format
    ):
        url = re.sub(pattern, replace, url)
    return url


def is_git_dir(wd: Union[str, Path]) -> bool:
    return _git_rev_parse("--git-dir", Path(wd)) is not None


def is_git_working_tree(wd: Union[str, Path]) -> bool:
    return _git_rev_parse("--show-toplevel", Path(wd)) is not None


def get_git_root(wd: Optional[Union[str, Path]] = None) -> Path:
    """
    Fetches the root of the git repo.
    This corresponds to the root directory in the case of a working tree,
    or the `.git/` directory in the case of a quarantine during pre-receive.

    :param wd: working directory, defaults to None
    :return: absolute path to the git root, as a string.
    """
    if wd is None:
        wd = Path.cwd()
    else:
        wd = Path(wd)
    check_git_dir(wd)
    top_level = _git_rev_parse(option="--show-toplevel", wd=wd)
    if top_level is not None:
        return Path(top_level)
    root = _git_rev_parse(option="--git-dir", wd=wd)
    if root is None:
        raise NotAGitDirectory()
    return Path(root).resolve()


def check_git_dir(wd: Optional[Union[str, Path]] = None) -> None:
    """Check if folder is git directory."""
    if wd is None:
        wd = Path.cwd()
    if not is_git_dir(wd):
        raise NotAGitDirectory()


def git(
    command: List[str],
    timeout: int = COMMAND_TIMEOUT,
    check: bool = True,
    cwd: Optional[Union[str, Path]] = None,
    log_stderr: bool = True,
    ignore_git_config: bool = True,
) -> str:
    """Calls git with the given arguments, returns stdout as a string"""
    env = os.environ.copy()
    # Ensure git messages are in English
    env["LANG"] = "C"
    # Ensure git behavior is not affected by the user git configuration, but give us a
    # way to set some configuration (useful for safe.directory)
    if ignore_git_config:
        env["GIT_CONFIG_GLOBAL"] = os.getenv("GG_GIT_CONFIG", "")
        env["GIT_CONFIG_SYSTEM"] = ""

    if cwd is None:
        cwd = Path.cwd()

    try:
        logger.debug("command=%s timeout=%d", command, timeout)
        result = subprocess.run(
            (
                [_get_git_path(), "-c", "core.quotePath=false"]
                + (
                    ["-c", "core.longpaths=true"]
                    if platform.system() == "Windows"
                    else []
                )
                + command
            ),
            check=check,
            capture_output=True,
            timeout=timeout,
            env=env,
            cwd=str(cwd),
        )
        if result.stderr and log_stderr:
            logger.warning(
                "command=%s, stderr=%s",
                command,
                result.stderr.decode("utf-8", errors="ignore"),
            )
        return result.stdout.decode("utf-8", errors="ignore").rstrip()
    except subprocess.CalledProcessError as exc:
        stderr = exc.stderr.decode("utf-8", errors="ignore")
        if log_stderr:
            logger.error("command=%s, stderr=%s", command, stderr)
        if "detected dubious ownership in repository" in stderr:
            raise GitError(
                "Git command failed because of a dubious ownership in repository.\n"
                "If you still want to run ggshield, make sure you mark "
                "the current repository as safe for git with:\n"
                "   git config --global --add safe.directory <YOUR_REPO>"
            )
        raise exc
    except subprocess.TimeoutExpired:
        raise GitCommandTimeoutExpired(
            'Command "{}" timed out'.format(" ".join(command))
        )


def git_ls(wd: Optional[Union[str, Path]] = None) -> List[str]:
    cmd = ["ls-files", "--recurse-submodules"]
    return git(cmd, timeout=600, cwd=wd).split("\n")


def _get_submodules_paths(wd: Optional[Union[str, Path]] = None) -> List[str]:
    cmd = ["submodule", "status"]
    return [
        submodule.split()[1]
        for submodule in git(cmd, timeout=600, cwd=wd).splitlines()
        if submodule
    ]


def git_ls_unstaged(wd: Optional[Union[str, Path]] = None) -> List[str]:
    # git command to get list of unstaged files in repo
    cmd = ["ls-files", "--others", "--exclude-standard"]
    unstaged_files = git(cmd, timeout=600, cwd=wd).splitlines()

    # --recurse-submodules is not compatible with --others so we
    # need to iterate over submodules to get the list of unstaged files
    for submodule in _get_submodules_paths(wd=wd):
        unstaged_files.extend(
            str(Path(submodule) / unstaged_file)
            for unstaged_file in git_ls_unstaged(
                wd=wd / Path(submodule) if wd else Path(submodule)
            )
        )
    return unstaged_files


def is_valid_git_commit_ref(ref: str, wd: Optional[Union[str, Path]] = None) -> bool:
    """
    Check if a reference is valid and can be resolved to a commit
    """
    if not wd:
        wd = Path.cwd()

    ref += "^{commit}"
    cmd = ["cat-file", "-e", ref]

    try:
        git(cmd, cwd=wd)
    except subprocess.CalledProcessError:
        return False

    return True


def check_git_ref(ref: str, wd: Optional[Union[str, Path]] = None) -> None:
    """Check if folder is a git repository and ref is a git reference."""
    if wd is None:
        wd = Path.cwd()
    check_git_dir(wd)

    if not is_valid_git_commit_ref(ref=ref, wd=wd):
        raise InvalidGitRefError(ref)


def get_list_commit_SHA(
    commit_range: str, max_count: Optional[int] = None
) -> List[str]:
    """
    Retrieve the list of commit SHA from a range.
    :param commit_range: A range of commits (ORIGIN...HEAD)
    :param max_count: If set, limits the number of SHA returned to this amount. This
    returns the *end* of the list, so max_count=3 returns [HEAD~2, HEAD~1, HEAD].
    """

    cmd = ["rev-list", "--reverse", *commit_range.split()]
    if max_count is not None:
        cmd.extend(["--max-count", str(max_count)])
    # Makes rev-list print "bad revision" instead of telling the range is ambiguous
    cmd.append("--")

    try:
        commit_list = git(cmd).split("\n")
    except subprocess.CalledProcessError as e:
        if b"bad revision" in e.stderr and "~1.." in commit_range:
            # We got asked to list commits for A~1...B. If A~1 does not exist, but A
            # does, then return A and its descendants until B.
            a_ref, remaining = commit_range.split("~1", maxsplit=1)
            if not is_valid_git_commit_ref(f"{a_ref}~1") and is_valid_git_commit_ref(
                a_ref
            ):
                commit_range = a_ref + remaining
                return [a_ref] + get_list_commit_SHA(commit_range)
        return []

    if "" in commit_list:
        commit_list.remove("")
        # only happens when git rev-list doesn't error
        # but returns an empty range, example git rev-list HEAD...

    return commit_list


def get_last_commit_sha_of_branch(branch_name: str) -> Optional[str]:
    """
    Returns the last commit sha of the given branch, or None
    if no commit could be found
    """
    # The branch is not directly available in CI env
    # We need to get commits through remotes
    last_target_commit = get_list_commit_SHA(branch_name, max_count=1)

    # Unable to find a commit on this branch
    # Consider it empty
    if not last_target_commit:
        return None

    return last_target_commit[0]


def get_repository_url_from_path(wd: Path) -> Optional[str]:
    """
    Returns one of the repository remote urls. Returns None if no remote are found,
    or the directory is not a repository.
    """
    remotes_raw: List[str] = []
    try:
        if not is_git_dir(wd):
            return None
        remotes_raw = git(["remote", "-v"], cwd=wd).splitlines()
    except (subprocess.CalledProcessError, OSError):
        return None

    url: Optional[str] = None
    for line in remotes_raw:
        if match := re.search(r"^(.*)\t(.*) \(fetch\)$", line):
            name, url = match.groups()
            if name == "origin":
                break
    return simplify_git_url(url) if url else None


def get_filepaths_from_ref(
    ref: str, wd: Optional[Union[str, Path]] = None
) -> List[Path]:
    """
    Fetches a list of all file paths indexed at a given reference in a git repository.
    :param ref: git reference, like a commit SHA, a relative reference like HEAD~1, ...
    :param wd: string path to the git repository. Defaults to current directory
    """
    if not wd:
        wd = Path.cwd()

    check_git_ref(ref, wd)

    filepaths = git(
        ["ls-tree", "--name-only", "--full-name", "-r", ref], cwd=wd
    ).splitlines()
    return [Path(path_str) for path_str in filepaths]


def get_staged_filepaths(wd: Optional[Union[str, Path]] = None) -> List[Path]:
    """
    Fetches a list of all file paths at the index in a git repository.
    :param wd: string path to the git repository. Defaults to current directory
    """
    if not wd:
        wd = Path.cwd()

    filepaths = git(["ls-files", "--full-name", "-c"], cwd=wd).splitlines()
    return [Path(path_str) for path_str in filepaths]


@lru_cache(None)
def read_git_file(ref: str, path: Path, wd: Optional[Union[str, Path]] = None) -> str:
    # Use as_posix to handle git and Windows
    return git(["show", f"{ref}:{path.as_posix()}"], cwd=wd)


def get_remotes(wd: Optional[Union[str, Path]] = None) -> List[str]:
    """List all configured git remotes."""
    if not wd:
        wd = Path.cwd()
    return git(["remote"], cwd=wd).splitlines()


def get_default_branch(wd: Optional[Union[str, Path]] = None) -> str:
    """
    Return the default branch of the repository.

    Try to get the default branch from a remote, either origin or the first remote,
    otherwise return the config init.defaultBranch.
    """
    if not wd:
        wd = Path.cwd()

    remotes = get_remotes(wd)
    remote = None
    if "origin" in remotes:
        remote = "origin"
    elif len(remotes) > 0:
        remote = remotes[0]

    default_branch = None
    if remote is not None:
        for line in git(["remote", "show", remote], cwd=wd).splitlines():
            line = line.strip()
            if line.startswith("HEAD branch: "):
                default_branch = remote + "/" + line[len("HEAD branch: ") :]
                break

    if default_branch is None:
        return git(["config", "init.defaultBranch"], cwd=wd).strip()

    return default_branch



from dataclasses import fields
from typing import Any, Dict, Optional

import marshmallow_dataclass
from marshmallow.decorators import pre_load
from pygitguardian.models_utils import FromDictMixin, ToDictMixin

from ggshield.core import ui


@marshmallow_dataclass.dataclass
class FilteredConfig(FromDictMixin, ToDictMixin):
    @classmethod
    @pre_load(pass_many=False)
    def filter_fields(cls, data: Dict[str, Any], **kwargs: Any) -> Dict[str, Any]:
        """
        Remove and alert on unknown fields.
        """
        field_names = {field_.name for field_ in fields(cls)}
        filtered_fields = {}
        for key, item in data.items():
            if key in field_names:
                filtered_fields[key] = item
            else:
                ui.display_warning(f"Unrecognized key in config: {key}")

        return filtered_fields


@marshmallow_dataclass.dataclass
class IgnoredMatch(FilteredConfig):
    match: str
    name: Optional[str] = None

    def __post_init__(self) -> None:
        if self.name is None:
            self.name = ""


IgnoredMatch.SCHEMA = marshmallow_dataclass.class_schema(IgnoredMatch)()



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_handle_scan_error[single file exception] 1'] = '''Error: Scanning failed. Results may be incomplete.
Error: Add the following files to your ignored_paths:
Error: - /home/user/too/long/file/name: filename:: [ErrorDetail(string='Ensure this field has no more than 256 characters.', code='max_length')]
'''

snapshots['test_handle_scan_error[too many documents] 1'] = '''Error: Scanning failed. Results may be incomplete.
Error: The following chunk is affected:
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
- /example
Error: 400:Too many documents to scan
'''



import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional

from ggshield.utils.git_shell import git

from .commit_utils import DIFF_EMPTY_COMMIT_INFO_BLOCK, HEADER_COMMON_ARGS, PatchHeader


_INFO_HEADER_REGEX = re.compile(
    r"Author:\s(?P<author>.*?) <(?P<email>.*?)>\nDate:\s+(?P<date>.+)"
)


@dataclass
class CommitInformation:
    author: str
    email: str
    date: str
    paths: List[Path]
    renames: Dict[Path, Path] = field(default_factory=dict)

    @staticmethod
    def from_patch_header(header_str: str) -> "CommitInformation":
        """
        Parse a patch header generated with `git show` or `git diff` using
        HEADER_COMMON_ARGS.

        Output format looks like this:

        ```
        commit $SHA
        Author: $NAME <$EMAIL>
        Date: $DATE

             $SUBJECT

             $BODY
        $RAW_FILE_LINES
        ```

        If the commit is empty the last line is absent and the header ends with `\n`.
        """
        header = PatchHeader.from_string(header_str)
        match = _INFO_HEADER_REGEX.search(header.info)
        assert match is not None, f"Failed to extract commit info from `{header.info}`"

        if header.files:
            # Usual commit, with files in it
            paths: List[Path] = []
            renames: Dict[Path, Path] = {}
            for file_info in header.files:
                paths.append(file_info.path)
                if file_info.old_path is not None:
                    renames[file_info.path] = file_info.old_path
        else:
            # Empty commit
            paths = []
            renames = {}

        return CommitInformation(**match.groupdict(), paths=paths, renames=renames)

    @staticmethod
    def from_staged(cwd: Optional[Path] = None) -> "CommitInformation":
        output = git(["diff", "--staged"] + HEADER_COMMON_ARGS, cwd=cwd)
        if not output:
            # This happens when there are no changes
            return CommitInformation("", "", "", [])

        return CommitInformation.from_patch_header(
            DIFF_EMPTY_COMMIT_INFO_BLOCK + output
        )

    @staticmethod
    def from_sha(sha: str, cwd: Optional[Path] = None) -> "CommitInformation":
        header = git(["show", sha] + HEADER_COMMON_ARGS, cwd=cwd)
        return CommitInformation.from_patch_header(header)



import json
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Optional, Union

import pytest
from requests import Response


def create_json_response(json_data: Dict[str, Any], status_code: int = 200) -> Response:
    """Helper function to create a Response returned by the requests package containing
    JSON content. This is required because `requests.Response` does not provide an easy
    way to build a response manually.

    Can be used on its own or with RequestMock."""
    response = Response()
    response.headers = {"content-type": "application/json"}
    response.status_code = status_code
    response._content = json.dumps(json_data).encode("utf-8")
    return response


def create_html_response(html: str, status_code: int = 200) -> Response:
    """Helper function to create a Response returned by the requests package containing
    HTML. This is required because `requests.Response` does not provide an easy
    way to build a response manually.

    Can be used on its own or with RequestMock."""
    response = Response()
    response.headers = {"content-type": "text/html"}
    response.status_code = status_code
    response._content = html.encode("utf-8")
    return response


# See ExpectedRequest.data_checker
DataChecker = Callable[[Any], None]
JSONChecker = Callable[[Union[List[str], Dict[str, Any]]], None]


@dataclass
class ExpectedRequest:
    """Stores an expected request and the response to return"""

    method: str
    endpoint: str
    response: Response

    # If defined, RequestMock will call this function with the value of the `data` field
    # it receives. It can be used to check the payload is as expected.
    data_checker: Optional[DataChecker] = None

    # If defined, RequestMock will call this function with the value of the `json` field
    # it receives. It can be used to check the payload is as expected.
    json_checker: Optional[JSONChecker] = None


class RequestMock:
    """
    Can be used to mock the `requests.Session.request()` method or the
    `requests.request()` from the requests package.

    Usage:

    ```python
    # Create the mock
    mock = RequestMock()

    # Install the mock
    monkeypatch.setattr("ggshield.core.client.Session.request", mock)

    # Add expected requests
    mock.add_GET("/foo1", create_json_response({"a": 12}))
    mock.add_POST("/login", create_json_response({"a": 12}))
    mock.add_GET("/not_found", create_json_response({"msg": "no-such-page"}, 400))

    # Execute the code expected to send requests
    # If a request does not match the expected requests, an assert will raise
    my_code()

    # Verify the tested code sent all the requests
    mock.assert_all_requests_happened()
    ```
    """

    def __init__(self):
        self._requests: List[ExpectedRequest] = []

    def add_GET(
        self,
        endpoint: str,
        response: Response,
    ):
        self.add_request(ExpectedRequest("GET", endpoint, response))

    def add_POST(
        self,
        endpoint: str,
        response: Response,
        data_checker: Optional[DataChecker] = None,
        json_checker: Optional[JSONChecker] = None,
    ):
        self.add_request(
            ExpectedRequest("POST", endpoint, response, data_checker, json_checker)
        )

    def add_request(self, request: ExpectedRequest) -> None:
        """Low-level method to add a request, it's simpler to use add_GET or add_POST
        instead"""
        self._requests.append(request)

    def assert_all_requests_happened(self) -> None:
        assert self._requests == []

    def __call__(
        self, method: str, url: str, data: Optional[Any] = None, **kwargs: Any
    ) -> Response:
        """This method is called by the tested code. It pops the next expected request
        and checks it matches with the received request."""
        method = method.upper()

        # Get the expected request
        if not self._requests:
            pytest.fail(f"Unexpected call: {method} {url}")
        request = self._requests.pop(0)

        # Check the current request matches the expected one
        assert url.endswith(
            request.endpoint
        ), f"Received call to this URL: {url} but expected a call to this endpoint: {request.endpoint}"
        assert method == request.method
        if request.data_checker:
            request.data_checker(data)
        if request.json_checker:
            request.json_checker(kwargs["json"])

        return request.response



import pytest

from tests.factories import get_line_index


TEST_CONTENT = """aaa
bb
cccc"""


@pytest.mark.parametrize(
    ("index", "expected_line_index"),
    (
        (1, 0),
        (4, 0),
        (5, 1),
        (7, 1),
        (8, 2),
        (11, 2),
    ),
)
def test_get_line_index(index, expected_line_index):
    assert get_line_index(TEST_CONTENT, index) == expected_line_index



import logging
import os
from pathlib import Path
from typing import Optional, Set

from dotenv import dotenv_values, load_dotenv

from ggshield.core import ui
from ggshield.utils.git_shell import get_git_root, is_git_dir
from ggshield.utils.os import getenv_bool


TRACKED_ENV_VARS = {
    "GITGUARDIAN_INSTANCE",
    "GITGUARDIAN_API_URL",
    "GITGUARDIAN_API_KEY",
}

logger = logging.getLogger(__name__)


def _find_dot_env() -> Optional[Path]:
    """Look for a .env to load, returns its path if found"""
    if env_var := os.getenv("GITGUARDIAN_DOTENV_PATH"):
        path = Path(env_var)
        if not path.is_file():
            ui.display_error(
                "GITGUARDIAN_DOTENV_PATH does not point to a valid .env file"
            )
            return None

        return path

    # Look for a .env in the current directory
    env = Path(".env")
    if env.is_file():
        return env

    # If we are in a git checkout, look for a .env at the root of the checkout
    if is_git_dir(os.getcwd()):
        env = get_git_root() / ".env"
        if env.is_file():
            return env

    return None


def load_dot_env() -> Set[str]:
    """
    Loads .env file into os.environ.
    Return the list of env vars that were set by the dotenv file
    among env vars in TRACKED_ENV_VARS
    """
    dont_load_env = getenv_bool("GITGUARDIAN_DONT_LOAD_ENV")
    if dont_load_env:
        logger.debug("Not loading .env, GITGUARDIAN_DONT_LOAD_ENV is set")
        return set()

    dot_env_path = _find_dot_env()
    if dot_env_path:
        dot_env_path = dot_env_path.absolute()
        logger.debug("Loading environment file %s", dot_env_path)
        load_dotenv(dot_env_path, override=True)

    return dotenv_values(dot_env_path).keys() & TRACKED_ENV_VARS



from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Set, Union, overload

import yaml
import yaml.parser
import yaml.scanner

from ggshield.core.constants import (
    AUTH_CONFIG_FILENAME,
    DEFAULT_CONFIG_FILENAME,
    USER_CONFIG_FILENAMES,
)
from ggshield.core.dirs import get_config_dir, get_project_root_dir, get_user_home_dir
from ggshield.core.errors import UnexpectedError
from ggshield.utils.git_shell import GitExecutableNotFound


def replace_dash_in_keys(data: Union[List[Any], Dict[str, Any]]) -> Set[str]:
    """Replace '-' with '_' in data keys.

    If a key exists in both dash and underscore versions, then only the underscore
    version is kept.

    Returns a set with the names of the renamed/removed dash keys."""
    dash_keys = set()

    if isinstance(data, dict):
        for key, value in list(data.items()):
            dash_keys.update(replace_dash_in_keys(value))
            if "-" in key:
                dash_value = data.pop(key)
                # If an underscore-version of the key exist, do not replace it
                new_key = key.replace("-", "_")
                if new_key not in data:
                    data[new_key] = dash_value
                dash_keys.add(key)
    elif isinstance(data, list):
        for element in data:
            dash_keys.update(replace_dash_in_keys(element))

    return dash_keys


def load_yaml_dict(path: Union[str, Path]) -> Optional[Dict[str, Any]]:
    path = Path(path)
    if not path.exists():
        return None

    with path.open() as f:
        try:
            data = yaml.safe_load(f) or {}
        except (yaml.parser.ParserError, yaml.scanner.ScannerError) as e:
            message = f"{path} is not a valid YAML file:\n{str(e)}"
            raise ValueError(message)

    if not isinstance(data, dict):
        raise ValueError(f"{path} should be a dictionary.")

    return data


def save_yaml_dict(
    data: Dict[str, Any], path: Union[str, Path], restricted: bool = False
) -> None:
    p = Path(path)
    p.parent.mkdir(parents=True, exist_ok=True)
    with p.open("w") as f:
        try:
            if restricted:
                # Restrict file permissions: read and write for owner only (600)
                p.chmod(0o600)

            stream = yaml.dump(data, indent=2, default_flow_style=False)
            f.write(stream)

        except Exception as e:
            raise UnexpectedError(f"Failed to save config to {path}:\n{str(e)}") from e


def get_auth_config_filepath() -> Path:
    return get_config_dir() / AUTH_CONFIG_FILENAME


def get_global_path(filename: str) -> Path:
    return get_user_home_dir() / filename


@overload
def find_global_config_path(*, to_write: Literal[False] = False) -> Optional[Path]: ...


@overload
def find_global_config_path(*, to_write: Literal[True]) -> Path: ...


def find_global_config_path(*, to_write: bool = False) -> Optional[Path]:
    """
    Returns the path to the user global config file (the file in the user home
    directory).
    If there is no such file:
    - If `to_write` is False, returns None.
    - If `to_write` is True, returns the path to the default file.

    This means the function never returns None if `to_write` is True.
    """
    for filename in USER_CONFIG_FILENAMES:
        path = get_global_path(filename)
        if path.exists():
            return path
    return get_global_path(DEFAULT_CONFIG_FILENAME) if to_write else None


def find_local_config_path() -> Optional[Path]:
    try:
        project_root_dir = get_project_root_dir(Path())
    except GitExecutableNotFound:
        project_root_dir = Path()
    for filename in USER_CONFIG_FILENAMES:
        path = project_root_dir / filename
        if path.exists():
            return path
    return None


def update_dict_from_other(dct: Dict[str, Any], other: Dict[str, Any]) -> None:
    """
    Merge values from `other` dict into `dct`. List entries are appended, sets are
    updated, dicts are merged.

    The merge happens in-place: `dct` is modified.
    """
    for key, value in other.items():
        try:
            if value is None:
                continue
            if isinstance(value, list):
                dct.setdefault(key, []).extend(value)
            elif isinstance(value, set):
                dct.setdefault(key, set()).update(value)
            elif isinstance(value, dict):
                update_dict_from_other(dct.setdefault(key, {}), value)
            else:
                dct[key] = value
        except AttributeError:
            raise UnexpectedError(f"Failed to load configuration on key '{key}'")


def remove_common_dict_items(
    dct: Dict[str, Any], reference_dct: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Returns a copy of `dct` with all items already in `reference_dct` removed.
    """

    result_dct = dict()
    for key, value in dct.items():
        reference_value = reference_dct[key]

        if isinstance(value, dict):
            value = remove_common_dict_items(value, reference_value)
            # Remove empty dicts
            if not value:
                continue
        else:
            if value == reference_value:
                continue

        result_dct[key] = value

    return result_dct


def remove_url_trailing_slash(url: str) -> str:
    if url[-1] == "/":
        return url[:-1]
    else:
        return url



from typing import Optional, Tuple
from unittest.mock import Mock

import pytest
from requests.exceptions import ConnectionError

from ggshield.__main__ import cli
from ggshield.core.config import Config
from ggshield.core.constants import DEFAULT_INSTANCE_URL
from ggshield.core.errors import ExitCode

from ..utils import add_instance_config


class TestAuthLogout:
    def test_logout_no_account_config(self, cli_fs_runner):
        """
        GIVEN -
        WHEN using the logout command and no token is saved in the configuration
        THEN the command exits with an explanatory message
        """
        instance_url = "https://dashboard.gitguardian.com"
        add_instance_config(with_account=False)
        exit_code, output = self.run_cmd(cli_fs_runner, instance_url)

        assert exit_code == ExitCode.UNEXPECTED_ERROR, output
        assert output == (
            f"Error: No token found for instance {instance_url}\n"
            "First try to login by running:\n"
            "  ggshield auth login\n"
        )

    @pytest.mark.parametrize(
        "instance_url", (DEFAULT_INSTANCE_URL, "https://some-gg-instance.com")
    )
    @pytest.mark.parametrize("revoke", (True, False))
    def test_valid_logout(self, revoke, instance_url, monkeypatch, cli_fs_runner):
        """
        GIVEN a saved instance configuration
        WHEN running the logout command
        THEN the specified instance data is erased
        AND the request for revocation is made if no flag was included
        """
        unrelated_url = "https://some-unrelated-gg-instance.com"

        post_mock = Mock(return_value=Mock(status_code=204, ok=True))
        monkeypatch.setattr("ggshield.core.client.GGClient.post", post_mock)

        token_name = "My great token"
        add_instance_config(instance_url=instance_url, token_name=token_name)

        # unrelated config that should remain unchanged
        add_instance_config(instance_url=unrelated_url)

        exit_code, output = self.run_cmd(cli_fs_runner, instance_url, revoke=revoke)

        if revoke:
            post_mock.assert_called_once()
        else:
            post_mock.assert_not_called()

        config = Config()
        assert config.auth_config.get_instance(instance_url).account is None
        assert (
            config.auth_config.get_instance(unrelated_url).account is not None
        ), "the unrelated instance should not be affected."

        assert exit_code == ExitCode.SUCCESS, output
        instance_url = instance_url or "https://dashboard.gitguardian.com"

        expected_output = f"Successfully logged out for instance {instance_url}\n\n"

        if revoke:
            expected_output += (
                "Your personal access token has been revoked and removed "
                "from your configuration.\n"
            )
        else:
            expected_output += (
                "Your personal access token has been removed "
                "from your configuration.\n"
            )

        assert output == expected_output

    def test_logout_revoke_timeout(self, monkeypatch, cli_fs_runner):
        """
        GIVEN a saved instance configuration
        WHEN running the logout command (with implied token revokation)
        AND the revoke request gets a timeout
        THEN the config remains unchanged
        AND the command exits with an explanatory message
        """

        post_mock = Mock(side_effect=ConnectionError("Http max retry"))
        monkeypatch.setattr("ggshield.core.client.GGClient.post", post_mock)

        add_instance_config()
        exit_code, output = self.run_cmd(cli_fs_runner)

        post_mock.assert_called_once()
        config = Config()
        assert config.auth_config.get_instance(DEFAULT_INSTANCE_URL).account is not None

        assert exit_code == ExitCode.UNEXPECTED_ERROR, output
        assert output == (
            "Error: Could not connect to GitGuardian.\n"
            "Please check your internet connection and if the specified URL is correct.\n"
        )

    def test_logout_server_error(self, monkeypatch, cli_fs_runner):
        """
        GIVEN a saved instance configuration
        WHEN running the logout command (with implied token revokation)
        AND the revoke request gets a server error response
        THEN the config remains unchanged
        AND the command exits with an explanatory message
        """
        post_mock = Mock(return_value=Mock(status_code=500, ok=False))
        monkeypatch.setattr("ggshield.core.client.GGClient.post", post_mock)

        add_instance_config()
        exit_code, output = self.run_cmd(cli_fs_runner)

        post_mock.assert_called_once()
        config = Config()
        assert config.auth_config.get_instance(DEFAULT_INSTANCE_URL).account is not None

        assert exit_code == ExitCode.AUTHENTICATION_ERROR, output
        assert output == (
            "Error: Could not perform the logout command "
            "because your token is already revoked or invalid.\n"
            "Please try with the following command:\n"
            "  ggshield auth logout --no-revoke\n"
        )

    def test_logout_all(self, monkeypatch, cli_fs_runner):
        """
        GIVEN several saved instances
        WHEN running the logout command with --all parameter
        THEN all tokens are revoked
        AND all account configs are deleted
        """
        post_mock = Mock(return_value=Mock(status_code=204, ok=True))
        monkeypatch.setattr("ggshield.core.client.GGClient.post", post_mock)

        for instance_url in [
            DEFAULT_INSTANCE_URL,
            "https://some-gg-instance.com",
            "https://some-other-gg-instance.com",
        ]:
            add_instance_config(instance_url)

        exit_code, output = self.run_cmd(cli_fs_runner, all_tokens=True)
        assert len(post_mock.call_args_list) == 3

        for instance in Config().auth_config.instances:
            assert instance.account is None, output

        assert exit_code == ExitCode.SUCCESS, output

    @staticmethod
    def run_cmd(
        cli_fs_runner,
        instance: Optional[str] = None,
        revoke: bool = True,
        all_tokens: bool = False,
    ) -> Tuple[int, str]:
        cmd = ["auth", "logout"]
        if instance is not None:
            cmd.append("--instance=" + instance)
        if not revoke:
            cmd.append("--no-revoke")
        if all_tokens:
            cmd.append("--all")
        result = cli_fs_runner.invoke(cli, cmd, color=False)
        return result.exit_code, result.output



import logging
import os
import sys
from typing import List, Optional, Tuple

from ggshield.core import ui
from ggshield.utils.git_shell import EMPTY_SHA, git, is_valid_git_commit_ref


logger = logging.getLogger(__name__)


OUTDATED_HOOK_MESSAGE = """The installed pre-push hook did not pass its command-line arguments to ggshield. This can cause the hook to fail if the name of the remote you are pushing to is not "origin".

This can happen if the hook has been created manually or by an old version of ggshield.

To fix it, either edit the hook manually or make a backup of it and reinstall it with the following command:

    ggshield install -m local -t pre-push -f
"""  # noqa: E501

BYPASS_MESSAGE = """  - if you use the pre-commit framework:

     SKIP=ggshield-push git push

  - otherwise (warning: the following command bypasses all pre-push hooks):

     git push --no-verify"""


def find_branch_start(commit: str, remote: str) -> Optional[str]:
    """
    Returns the first local-only commit of the branch.
    Returns None if the branch does not contain any new commit.
    """
    # List all ancestors of `commit` which are not in `remote`
    # Based on _pre_push_ns() from pre-commit
    #
    # Note: The `--remotes` argument MUST be set using a `=`: `--remotes={remote}` works,
    # but `--remotes {remote}` fails.
    output = git(
        [
            "rev-list",
            commit,
            "--topo-order",
            "--reverse",
            "--not",
            f"--remotes={remote}",
        ]
    )
    ancestors = output.splitlines()

    if ancestors:
        return ancestors[0]
    return None


def collect_commits_from_stdin(remote_name: str) -> Tuple[str, str]:
    """
    Collect pre-commit variables from stdin.
    """
    prepush_input = sys.stdin.read().strip()
    logger.debug("input=%s", prepush_input)
    if not prepush_input:
        # Happens when there's nothing to push
        return (EMPTY_SHA, EMPTY_SHA)

    # TODO There can be more than one line here, for example when pushing multiple
    # branches. We should support this.
    line = prepush_input.splitlines()[0]
    _, local_commit, _, remote_commit = line.split(maxsplit=3)

    if is_valid_git_commit_ref(remote_commit):
        # Pushing to an existing branch
        return (local_commit, remote_commit)

    # Pushing to a new branch
    start_commit = find_branch_start(local_commit, remote_name)
    if start_commit is None:
        return local_commit, local_commit
    return (local_commit, f"{start_commit}~1")


def collect_commits_from_precommit_env() -> Tuple[Optional[str], Optional[str]]:
    """
    Collect from pre-commit framework environment.
    """
    # pre-commit framework <2.2.0
    local_commit = os.getenv("PRE_COMMIT_SOURCE", None)
    remote_commit = os.getenv("PRE_COMMIT_ORIGIN", None)

    if local_commit is None or remote_commit is None:
        # pre-commit framework >=2.2.0
        local_commit = os.getenv("PRE_COMMIT_FROM_REF", None)
        remote_commit = os.getenv("PRE_COMMIT_TO_REF", None)

    return (local_commit, remote_commit)


def collect_commits_refs(prepush_args: List[str] = []) -> Tuple[str, str]:
    local_commit, remote_commit = collect_commits_from_precommit_env()
    if local_commit is None or remote_commit is None:
        if len(prepush_args) == 0:
            ui.display_warning(OUTDATED_HOOK_MESSAGE)
            remote_name = "origin"
        else:
            remote_name = prepush_args[0]
        local_commit, remote_commit = collect_commits_from_stdin(remote_name)
    return (local_commit, remote_commit)



import logging
import os
import sys
from typing import Optional, Tuple

from ggshield.core import ui
from ggshield.core.errors import UnexpectedError
from ggshield.utils.git_shell import EMPTY_SHA, git
from ggshield.utils.os import getenv_float, getenv_int


logger = logging.getLogger(__name__)


# GitHub timeouts every pre-receive hook after 5s with an error.
# We try and anticipate that so we can control the return code
PRERECEIVE_TIMEOUT = 4.5


BYPASS_MESSAGE = """\n     git push -o breakglass"""


def get_prereceive_timeout() -> float:
    try:
        return getenv_float("GITGUARDIAN_TIMEOUT", PRERECEIVE_TIMEOUT)
    except BaseException as e:
        ui.display_error(f"Unable to parse GITGUARDIAN_TIMEOUT: {str(e)}")
        return PRERECEIVE_TIMEOUT


def get_breakglass_option() -> bool:
    """Test all options passed to git for `breakglass`"""
    option_count = getenv_int("GIT_PUSH_OPTION_COUNT")
    if option_count is not None:
        for option in range(option_count):
            if os.getenv(f"GIT_PUSH_OPTION_{option}", "") == "breakglass":
                ui.display_info(
                    "SKIP: breakglass detected. Skipping GitGuardian pre-receive hook."
                )
                return True

    return False


def find_branch_start(commit: str) -> Optional[str]:
    """
    Returns the first local-only commit of the branch.
    Returns None if the branch does not contain any new commit.
    """
    # List all ancestors of `commit` which are not in any branches
    output = git(
        ["rev-list", commit, "--topo-order", "--reverse", "--not", "--branches"]
    )
    ancestors = output.splitlines()

    if ancestors:
        return ancestors[0]
    return None


def parse_stdin() -> Optional[Tuple[str, str]]:
    """
    Parse stdin and return the first and last commit to scan,
    or None if there is nothing to do, allowing for early stopping.
    """
    prereceive_input = sys.stdin.read().strip()
    if not prereceive_input:
        raise UnexpectedError(f"Invalid input arguments: '{prereceive_input}'")

    # TODO There can be more than one line here, for example when pushing multiple
    # branches. We should support this.
    line = prereceive_input.splitlines()[0]
    logger.debug("stdin: %s", line)
    _old_commit, new_commit, _ = line.split(maxsplit=2)

    if new_commit == EMPTY_SHA:
        # Deletion event, nothing to do
        ui.display_info("Deletion event or nothing to scan.")
        return None

    # ignore _old_commit because in case of a force-push, it is going to be overwritten
    # and should not be scanned (see #437)
    start_commit = find_branch_start(new_commit)
    if start_commit is None:
        # branch does not contain any new commit
        old_commit = new_commit
    else:
        old_commit = f"{start_commit}~1"

    assert old_commit != EMPTY_SHA
    assert new_commit != EMPTY_SHA
    if old_commit == new_commit:
        ui.display_info("Pushed branch does not contain any new commit.")
        return None

    return (old_commit, new_commit)



from typing import Any, Optional

import click

from ggshield.cmd.config.constants import FIELD_NAMES, FIELD_NAMES_DOC, FIELDS
from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.errors import UnknownInstanceError


@click.command(
    help=f"""
Print the value of the given configuration key.
If `--instance` is passed, retrieve the value for this specific instance.

{FIELD_NAMES_DOC}"""
)
@click.argument(
    "field_name",
    nargs=1,
    type=click.Choice(FIELD_NAMES),
    required=True,
    metavar="KEY",
)
@click.option(
    "--instance",
    "instance_url",
    required=False,
    type=str,
    metavar="URL",
    help="Get per instance configuration.",
)
@add_common_options()
@click.pass_context
def config_get_cmd(
    ctx: click.Context, field_name: str, instance_url: Optional[str], **kwargs: Any
) -> int:
    config = ContextObj.get(ctx).config
    field = FIELDS[field_name]

    if field.auth_config:
        if instance_url:
            instance_config = config.auth_config.get_instance(instance_url)
            value = getattr(instance_config, field_name, None)
        else:
            value = getattr(config.auth_config, field_name, None)
            if value is None and field.per_instance_ok:
                # No value, try to get the one from the default instance
                try:
                    instance_config = config.auth_config.get_instance(
                        config.instance_name
                    )
                except UnknownInstanceError:
                    pass
                else:
                    value = getattr(instance_config, field_name, None)
    else:
        value = getattr(config.user_config, field_name, None)

    if value is None:
        value = "not set"
    click.echo(f"{field_name}: {value}")
    return 0



import os
from typing import Type
from unittest.mock import Mock, patch

import click
import pytest
import requests.exceptions
from pygitguardian import GGClient
from pygitguardian.models import Detail

from ggshield.core.client import check_client_api_key, create_client_from_config
from ggshield.core.config import Config
from ggshield.core.errors import (
    APIKeyCheckError,
    ServiceUnavailableError,
    UnexpectedError,
    UnknownInstanceError,
)


@pytest.mark.parametrize(
    ("response", "error_class"),
    (
        (Detail("Guru Meditation", 500), ServiceUnavailableError),
        (Detail("Nobody here", 404), UnexpectedError),
        (Detail("Unauthorized", 401), APIKeyCheckError),
    ),
)
def test_check_client_api_key_error(response: Detail, error_class: Type[Exception]):
    """
    GIVEN a client returning an error when its healthcheck endpoint is called
    WHEN check_client_api_key() is called
    THEN it raises the appropriate exception
    """
    client_mock = Mock(spec=GGClient)
    client_mock.base_uri = "http://localhost"
    client_mock.read_metadata.return_value = response
    with pytest.raises(error_class):
        check_client_api_key(client_mock)


def test_check_client_api_key_network_error():
    """
    GIVEN a client with a wrong instance URL
    WHEN check_client_api_key() is called
    THEN it raises an UnexpectedError
    """
    client_mock = Mock()
    client_mock.health_check = Mock(side_effect=requests.exceptions.ConnectionError)
    client_mock.read_metadata = Mock(return_value=Detail("Not found", 404))
    with pytest.raises(UnexpectedError):
        check_client_api_key(client_mock)


def test_retrieve_client_invalid_api_url():
    """
    GIVEN a GITGUARDIAN_API_URL missing its https scheme
    WHEN retrieve_client() is called
    THEN it raises a UsageError
    """
    url = "no-scheme.com"
    environ = os.environ.copy()
    environ.pop("GITGUARDIAN_INSTANCE", None)
    environ["GITGUARDIAN_API_URL"] = url

    with pytest.raises(
        click.UsageError,
        match=f"Invalid scheme for API URL '{url}', expected HTTPS",
    ):
        with patch.dict(os.environ, environ, clear=True):
            create_client_from_config(Config())


def test_retrieve_client_invalid_api_key():
    """
    GIVEN a GITGUARDIAN_API_KEY with a non-latin-1 character
    WHEN retrieve_client() is called
    THEN it raises a UnexpectedError
    """
    with pytest.raises(UnexpectedError, match="Invalid value for API Key"):
        with patch.dict(os.environ, {"GITGUARDIAN_API_KEY": "\u2023"}):
            create_client_from_config(Config())


def test_retrieve_client_blank_state(isolated_fs):
    """
    GIVEN a blank state (no config, no environment variable)
    WHEN retrieve_client() is called
    THEN the exception message is user-friendly for new users
    """
    with pytest.raises(
        APIKeyCheckError,
        match="A GitGuardian API key is needed to use ggshield.",
    ):
        with patch.dict(os.environ, clear=True):
            create_client_from_config(Config())


def test_retrieve_client_unknown_custom_dashboard_url(isolated_fs):
    """
    GIVEN an auth config telling the client to use a custom instance
    WHEN retrieve_client() is called
    AND the custom instance does not exist
    THEN the exception message mentions the instance name
    """
    with pytest.raises(
        UnknownInstanceError,
        match="Unknown instance: 'https://example.com'",
    ):
        with patch.dict(os.environ, clear=True):
            config = Config()
            config.cmdline_instance_name = "https://example.com"
            create_client_from_config(config)



import random

import factory
import factory.fuzzy
from pygitguardian.models import Match, PolicyBreak, ScanResult

from ggshield.core.scan.scannable import StringScannable
from ggshield.utils.git_shell import Filemode
from ggshield.verticals.secret.secret_scan_collection import Secret
from tests.factory_constants import DETECTOR_NAMES, MATCH_NAMES


def get_line_index(content, index):
    """Return the index of the line containing the character at the given index"""
    current_line_index = 0
    lines = content.splitlines(keepends=True)
    while True:
        line = lines.pop(0)
        if index <= len(line):
            return current_line_index
        index -= len(line)
        current_line_index += 1


class ScannableFactory(factory.Factory):
    class Meta:
        model = StringScannable

    url = factory.Faker("hostname")
    content = factory.Faker("text")
    # Only returning FILE for new, since diff would need a custom content
    filemode = Filemode.FILE


class MatchFactory(factory.Factory):
    class Meta:
        model = Match

    content = factory.Faker("text")
    match_len = factory.fuzzy.FuzzyInteger(5, 15)
    index_start = factory.lazy_attribute(
        lambda obj: random.randint(0, len(obj.content) - obj.match_len)
    )
    index_end = factory.lazy_attribute(lambda obj: obj.index_start + obj.match_len)
    match = factory.lazy_attribute(
        lambda obj: obj.content[obj.index_start : obj.index_end]
    )
    line_start = factory.lazy_attribute(
        lambda obj: get_line_index(obj.content, obj.index_start)
    )
    line_end = factory.lazy_attribute(
        lambda obj: get_line_index(obj.content, obj.index_end)
    )
    match_type = factory.lazy_attribute(lambda obj: random.choice(MATCH_NAMES))


class PolicyBreakFactory(factory.Factory):
    class Meta:
        model = PolicyBreak

    break_type = factory.lazy_attribute(lambda obj: random.choice(DETECTOR_NAMES))
    policy = "Secrets detection"
    validity = "valid"
    known_secret = False
    incident_url = None
    is_excluded = False
    exclude_reason = None
    diff_kind = None
    content = factory.Faker("text")
    nb_matches = factory.fuzzy.FuzzyInteger(1, 2)

    @factory.lazy_attribute
    def matches(self):
        # Note: matches may overlap, but at least we ensure they
        # have different names
        match_names = random.sample(MATCH_NAMES, self.nb_matches)
        return [
            MatchFactory(match_type=match_name, content=self.content)
            for match_name in match_names
        ]


class ScanResultFactory(factory.Factory):
    class Meta:
        model = ScanResult

    policy_break_count = factory.lazy_attribute(lambda obj: len(obj.policy_breaks))
    policy_breaks = []
    policies = ["Secrets detection"]
    is_diff = False


class SecretFactory(factory.Factory):
    class Meta:
        model = Secret

    detector = factory.lazy_attribute(lambda obj: random.choice(DETECTOR_NAMES))
    validity = "valid"
    known_secret = True
    incident_url = None
    matches = []
    ignore_reason = None
    diff_kind = None



import tarfile
from pathlib import Path

from tests.functional.utils import run_ggshield_scan


CURRENT_DIR = Path(__file__).parent


def test_scan_archive(tmp_path: Path) -> None:
    archive_path = tmp_path / "test.tar.gz"

    with tarfile.open(archive_path, "w:gz") as tar:
        for path in CURRENT_DIR.rglob("*"):
            tar.add(path)
    assert archive_path.exists()

    run_ggshield_scan("archive", str(archive_path), expected_code=0)



DETECTOR_NAMES = [
    "Basic Auth String",
    "Generic Password",
    "JSON Web Token",
    "Generic Terraform Variable Secret",
    "Generic Database Assignment",
    "Company Email Password",
    "Generic Password",
    "Base64 Generic High Entropy Secret",
    "Generic Database Assignment",
    "Generic CLI Option Secret",
    "Username Password",
    "Generic High Entropy Secret",
    "Base64 Basic Authentication",
    "Bearer Token",
    "Authentication Tuple",
    "Typeform API Token",
    "New Relic API Key",
    "Pingdom token v3",
    "Datadog Keys",
    "Databricks Authentication Token With Hostname",
    "GitGuardian Public Monitoring API Key",
    "GitHub Server-to-server Token",
    "Infracost API Key",
    "Facebook App Keys",
    "Firebase Cloud Messaging API Key",
    "Intercom Token",
    "Sourcegraph Access Token v1",
    "Stripe Webhook Secret",
    "GitLab Token",
    "New Relic API Service Key",
    "Eventbrite OAuth2 Token",
    "Base64 AWS SES Keys",
    "Doppler API Key",
    "Heartland API key",
    "Tailscale Pre-Authentication Key",
    "Kraken Keys",
    "Coveralls Repository Token",
    "Docker Credentials",
    "Algolia Monitoring Keys",
    "Grafana Token",
    "PackageCloud API Token",
    "Square Access Token",
    "DigitalOcean Token",
    "Sourcegraph Access Token v3",
    "Akamai API Credentials",
    "Linode Personal Access Token",
    "Scalr API Access Token",
    "FullContact Key",
    "Nylas API Key",
    "Plaid Access Token",
]
MATCH_NAMES = [
    "apikey",
    "client_id",
    "client_secret",
    "host",
    "password",
    "username",
    "token",
    "port",
    "scheme",
    "connection_uri",
    "subdomain",
    "private_key",
    "domain",
    "secret_key",
    "access_token",
    "project_id",
    "cloud_name",
    "database",
    "client_token",
    "secret_token",
    "tenant_id",
    "private_key_id",
    "integration_key",
    "azure_endpoint",
    "app_id",
    "cluster",
    "pub_key",
    "sub_key",
    "environment",
    "refresh_token",
    "organization",
    "session_token",
    "connection_string",
    "account",
    "user",
    "client_certificate",
    "client_key",
    "config_value",
]



from .client import HASH_REGEX, PREFIX_LENGTH, PREFIX_REGEX, HMSLClient, Match, Secret
from .utils import get_client


__all__ = [
    "HASH_REGEX",
    "PREFIX_REGEX",
    "PREFIX_LENGTH",
    "Match",
    "Secret",
    "HMSLClient",
    "get_client",
]



from unittest.mock import patch

import pytest

from ggshield.cmd.utils.hooks import check_user_requested_skip


@pytest.mark.parametrize("env_var_value", ["ggshield", "ggshield,foo", "ggshield, foo"])
def test_check_user_requested_skip_skipped(env_var_value):
    with patch.dict("os.environ", {"SKIP": env_var_value}, clear=True):
        assert check_user_requested_skip()


@pytest.mark.parametrize("env_var_value", ["", "foo", "foo,bar"])
def test_check_user_requested_skip_not_skipped(env_var_value):
    with patch.dict("os.environ", {"SKIP": env_var_value}, clear=True):
        assert not check_user_requested_skip()


def test_check_user_requested_skip_no_var():
    with patch.dict("os.environ", {}, clear=True):
        assert not check_user_requested_skip()



from typing import Any, Dict, List, Optional, Union

from ggshield.core.url_utils import api_to_dashboard_url


def convert_v1_config_dict(
    data: Dict[str, Any], deprecation_messages: List[str]
) -> Dict[str, Any]:
    """
    Takes a dict representing a v1 .gitguardian.yaml and returns a v2 dict.
    Appends any deprecation message to `deprecation_messages`.
    """

    # If data contains the old "api-url" key, turn it into an "instance" key,
    # but only if there is no "instance" key
    try:
        api_url = data.pop("api_url")
    except KeyError:
        pass
    else:
        if "instance" not in data:
            data["instance"] = api_to_dashboard_url(api_url, warn=True)

    if "all_policies" in data:
        deprecation_messages.append(
            "The `all_policies` option has been deprecated and is now ignored."
        )

    if "ignore_default_excludes" in data:
        deprecation_messages.append(
            "The `ignore_default_exclude` option has been deprecated and is now ignored."
        )

    secret_dct = {}
    if matches_ignore := data.get("matches_ignore"):
        secret_dct["ignored_matches"] = [
            _convert_matches_ignore_entry(x) for x in matches_ignore
        ]

    def copy_if_set(dct: Dict[str, Any], dst_key: str, src_key: Optional[str] = None):
        """Helper function: if `src_key` exists in `data`, copies its value to
        `dct[dst_key]`.
        If `src_key` is None, use `dst_key` as the source key.
        """
        if src_key is None:
            src_key = dst_key
        try:
            value = data[src_key]
        except KeyError:
            return
        dct[dst_key] = value

    copy_if_set(secret_dct, "show_secrets")
    copy_if_set(secret_dct, "ignored_detectors", "banlisted_detectors")
    copy_if_set(secret_dct, "ignored_paths", "paths_ignore")

    dct = {
        "secret": secret_dct,
    }
    copy_if_set(dct, "instance")
    copy_if_set(dct, "exit_zero")
    copy_if_set(dct, "verbose")
    copy_if_set(dct, "allow_self_signed")
    copy_if_set(dct, "max_commits_for_hook")

    return dct


def _convert_matches_ignore_entry(entry: Union[str, Dict[str, Any]]) -> Dict[str, Any]:
    """
    v1 config format allowed to use just a hash of the secret for matches_ignore
    field v2 does not. This function converts the hash-only entry.
    """
    if isinstance(entry, str):
        return {"name": "", "match": entry}
    else:
        return entry



from pathlib import Path

from tests.functional.utils import run_ggshield_scan


CURRENT_DIR = Path(__file__).parent


def test_scan_no_api_key(monkeypatch) -> None:
    monkeypatch.delenv("GITGUARDIAN_API_KEY", raising=False)
    run_ggshield_scan("path", __file__, cwd=CURRENT_DIR, expected_code=3)


def test_scan_invalid_api_key(monkeypatch) -> None:
    monkeypatch.setenv("GITGUARDIAN_API_KEY", "not_a_valid_key")
    run_ggshield_scan("path", __file__, cwd=CURRENT_DIR, expected_code=3)



from typing import Any

import click

from ggshield.cmd.honeytoken.create import create_cmd
from ggshield.cmd.honeytoken.create_with_context import create_with_context_cmd
from ggshield.cmd.utils.common_options import add_common_options


@click.group(
    commands={"create": create_cmd, "create-with-context": create_with_context_cmd}
)
@add_common_options()
def honeytoken_group(**kwargs: Any) -> None:
    """Commands to work with honeytokens."""



from datetime import datetime
from typing import Optional

from ggshield.core.config import AccountConfig, Config, InstanceConfig
from ggshield.core.constants import DEFAULT_INSTANCE_URL


def add_instance_config(
    instance_url: str = DEFAULT_INSTANCE_URL,
    token_name: str = "some token name",
    expiry_date: Optional[datetime] = None,
    with_account: Optional[bool] = True,
    default_token_lifetime: Optional[int] = None,
):
    """
    Creates an InstanceConfig with the provided arguments and adds it to the config
    """
    if with_account:
        account_config = AccountConfig(
            workspace_id=1,
            token="some token",
            expire_at=expiry_date,
            token_name=token_name,
            type="",
        )
    else:
        account_config = None

    instance_config = InstanceConfig(
        account=account_config,
        url=instance_url,
        default_token_lifetime=default_token_lifetime,
    )
    config = Config()
    config.auth_config.instances.append(instance_config)
    config.save()



import codecs
import logging
import urllib.parse
from abc import ABC, abstractmethod
from io import SEEK_END, SEEK_SET
from pathlib import Path
from typing import BinaryIO, Optional, Tuple

import charset_normalizer
from charset_normalizer import CharsetMatch

from ggshield.utils.git_shell import Filemode


logger = logging.getLogger(__name__)


# Our worse encoding (UTF-32) would take 4 bytes to encode ASCII, where UTF-8 would take
# only 1. If the file is longer than byte_size / UTF8_TO_WORSE_OTHER_ENCODING_RATIO, no
# need to look into it: it's too big.
UTF8_TO_WORSE_OTHER_ENCODING_RATIO = 4


class DecodeError(Exception):
    """
    Raised when a Scannable cannot determine the encoding of its content.

    Similar to UnicodeDecodeError, but easier to instantiate.
    """

    pass


class Scannable(ABC):
    """Base class for content that can be scanned by GGShield"""

    def __init__(self, filemode: Filemode = Filemode.FILE):
        self.filemode = filemode
        self._content: Optional[str] = None
        self._utf8_encoded_size: Optional[int] = None

    @property
    @abstractmethod
    def url(self) -> str:
        """Act as a unique identifier for the Scannable. May use custom protocols if
        required."""
        raise NotImplementedError

    @property
    @abstractmethod
    def filename(self) -> str:
        """To avoid breakage with the rest of the code base, implementations currently
        return the URL or path of the instance for now, but it should really return
        just the filename, or be removed."""
        # TODO: make this really return the filename, or remove it
        raise NotImplementedError

    @property
    @abstractmethod
    def path(self) -> Path:
        raise NotImplementedError

    @abstractmethod
    def is_longer_than(self, max_utf8_encoded_size: int) -> bool:
        """Return true if the length of the *utf-8 encoded* content is greater than
        `max_utf8_encoded_size`.
        When possible, implementations must try to answer this without reading all
        content.
        Raise `DecodeError` if the content cannot be decoded.
        """
        raise NotImplementedError

    @abstractmethod
    def _read_content(self) -> None:
        """Read the content of the scannable  if necessary, store it in `self._content`
        and the UTF8 encoded size in `self._utf8_encoded_size`"""
        raise NotImplementedError

    @property
    def content(self) -> str:
        """Return the decoded content of the scannable"""
        if self._content is None:
            self._read_content()
            if self._content is None:
                raise ValueError("content is None after reading")
        return self._content

    @property
    def utf8_encoded_size(self) -> int:
        if self._utf8_encoded_size is None:
            self._read_content()
            if self._utf8_encoded_size is None:
                raise ValueError("utf8_encoded_size is None after reading")
        return self._utf8_encoded_size

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} url={self.url} filemode={self.filemode}>"

    @staticmethod
    def _decode_bytes(
        raw_document: bytes, charset_match: Optional[CharsetMatch] = None
    ) -> Tuple[str, int]:
        """Low level helper function to decode bytes using `charset_match`. If
        `charset_match` is not provided, tries to determine it itself.

        Returns a tuple of (decoded_content, utf8_encoded_size).

        Raises DecodeError if the document cannot be decoded."""
        if charset_match is None:
            charset_match = charset_normalizer.from_bytes(raw_document).best()
            if charset_match is None:
                # This means we were not able to detect the encoding
                raise DecodeError

        # Special case for utf_8 + BOM: `bytes.decode()` does not skip the BOM, so do it
        # ourselves
        if charset_match.encoding == "utf_8" and raw_document.startswith(
            codecs.BOM_UTF8
        ):
            raw_document = raw_document[len(codecs.BOM_UTF8) :]
        content = raw_document.decode(charset_match.encoding, errors="replace")

        if charset_match.encoding in {"utf_8", "ascii"}:
            # The document is already in UTF-8, no need to encode it as UTF-8 to
            # determine UTF-8 encoded size.
            utf8_encoded_size = len(raw_document)
        else:
            utf8_encoded_size = len(content.encode(errors="replace"))

        return content, utf8_encoded_size

    @staticmethod
    def _is_file_longer_than(
        fp: BinaryIO, max_utf8_encoded_size: int
    ) -> Tuple[bool, Optional[str], Optional[int]]:
        """Helper function to implement is_longer_than() for file-based Scannable classes.

        Returns a tuple of:
        - True if file is longer than `size`, False otherwise
        - The decoded content as a string if the file has been fully read, None otherwise
        - The utf8-encoded size if we know it, None otherwise

        Raises DecodeError if the file cannot be decoded.
        """
        # Get the byte size
        assert fp.seekable()
        byte_size = fp.seek(0, SEEK_END)

        if byte_size > max_utf8_encoded_size * UTF8_TO_WORSE_OTHER_ENCODING_RATIO:
            # Even if the file used the worst encoding (UTF-32), encoding the content of
            # this file as UTF-8 would produce a file longer than
            # `max_utf8_encoded_size`, so bail out
            return True, None, None

        # Determine the encoding
        fp.seek(0, SEEK_SET)
        charset_matches = charset_normalizer.from_fp(fp)
        charset_match = charset_matches.best()
        if charset_match is None:
            raise DecodeError

        logger.debug('filename="%s" charset=%s', fp.name, charset_match.encoding)
        if charset_match.encoding in {"utf_8", "ascii"}:
            # Shortcut: the content is already in UTF-8 (or ASCII, which is a subset of
            # utf-8), no need to decode anything
            return byte_size > max_utf8_encoded_size, None, byte_size

        # We can't know if the file is longer without reading its content, do it now
        fp.seek(0, SEEK_SET)
        content, utf8_encoded_size = Scannable._decode_bytes(fp.read(), charset_match)
        logger.debug('filename="%s" utf8_encoded_size=%d', fp.name, utf8_encoded_size)
        if utf8_encoded_size > max_utf8_encoded_size:
            return True, None, utf8_encoded_size
        else:
            # We read the whole file, keep it
            return False, content, utf8_encoded_size


class StringScannable(Scannable):
    """Implementation of Scannable for content already loaded in memory"""

    def __init__(self, url: str, content: str, filemode: Filemode = Filemode.FILE):
        super().__init__(filemode)
        self._url = url
        self._path: Optional[Path] = None
        self._content = content

    def _read_content(self) -> None:
        assert self._content is not None
        if self._utf8_encoded_size is None:
            self._utf8_encoded_size = len(self._content.encode(errors="replace"))

    @property
    def url(self) -> str:
        return self._url

    @property
    def filename(self) -> str:
        return str(self._url)

    @property
    def path(self) -> Path:
        if self._path is None:
            result = urllib.parse.urlparse(self._url)
            self._path = Path(result.path)
        return self._path

    def is_longer_than(self, max_utf8_encoded_size: int) -> bool:
        self._read_content()
        assert self._utf8_encoded_size is not None
        return self._utf8_encoded_size > max_utf8_encoded_size



import logging
import sys
from typing import Any

import rich.markup
from rich.console import Console
from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn
from typing_extensions import Self

from ..ggshield_ui import NAME_BY_LEVEL, DebugInfo, GGShieldProgress, GGShieldUI, Level


COLOR_BY_LEVEL = {
    Level.DEBUG: "green",
    Level.VERBOSE: "white",
    Level.INFO: "blue",
    Level.WARNING: "yellow",
    Level.ERROR: "red",
}

LEVEL_BY_LOGGING_LEVEL = {
    logging.DEBUG: Level.DEBUG,
    logging.INFO: Level.INFO,
    logging.WARNING: Level.WARNING,
    logging.ERROR: Level.ERROR,
}


class RichGGShieldProgress(GGShieldProgress):
    def __init__(self, console: Console, total: int) -> None:
        super().__init__()

        self.progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TextColumn("{task.completed} / {task.total}"),
            console=console,
        )
        self.task = self.progress.add_task("Scanning...", total=total)

    def advance(self, amount: int) -> None:
        self.progress.advance(self.task, advance=amount)

    def __enter__(self) -> Self:
        self.progress.__enter__()
        return self

    def __exit__(self, *args: Any) -> None:
        self.progress.__exit__(*args)


class RichGGShieldUI(GGShieldUI):
    """
    Implementation of GGShieldUI using rich, for a more user-friendly terminal output.
    """

    def __init__(self):
        super().__init__()
        self.console = Console(file=sys.stderr)
        self._previous_timestamp = ""

    def create_progress(self, total: int) -> GGShieldProgress:
        return RichGGShieldProgress(self.console, total)

    def _echo(self, level: Level, message: str) -> None:
        message = rich.markup.escape(message)

        if self.level == Level.DEBUG:
            self._debug_echo(level, message)
        else:
            self._normal_echo(level, message)

    def _debug_echo(self, level: Level, rich_message: str) -> None:
        color = COLOR_BY_LEVEL[level]
        name = rich.markup.escape(f"[{NAME_BY_LEVEL[level][0]}]")
        message = f"[{color}]{name}[/] {rich_message}"

        self.console.print(self._prefix() + message)

    def _normal_echo(self, level: Level, message: str) -> None:
        if level <= Level.WARNING:
            color = COLOR_BY_LEVEL[level]
            name = NAME_BY_LEVEL[level]
            message = f"[{color}]{name}:[/] {message}"

        self.console.print(
            self._prefix() + message,
            # Disable highlight if we are not debugging, this way we get more control
            # on the output.
            highlight=False,
        )

    def _echo_heading(self, message: str) -> None:
        message = rich.markup.escape(message)
        prefix = self._prefix()
        self.console.print(f"\n[green]{prefix}# {message}[/]", highlight=False)

    def _prefix(self) -> str:
        if self.level < Level.DEBUG:
            return ""
        info = DebugInfo.create()

        # Do not repeat timestamp if it's the same as before
        if info.timestamp == self._previous_timestamp:
            timestamp = " " * len(info.timestamp)
        else:
            timestamp = info.timestamp
            self._previous_timestamp = timestamp

        return (
            f"[cyan]{timestamp}[/] [bright_black]{info.process_id}:{info.thread_id}[/] "
        )

    def log(self, record: logging.LogRecord) -> None:
        level = LEVEL_BY_LOGGING_LEVEL.get(record.levelno, Level.ERROR)
        msg = f"[magenta]{record.name}:{record.lineno}[/] {record.getMessage()}"
        self._debug_echo(level, msg)



import platform
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Union

from ggshield import __version__
from ggshield.core.git_hooks.ci.repository import get_repository_url_from_ci
from ggshield.utils.git_shell import get_repository_url_from_path
from ggshield.utils.os import get_os_info

from .scan_mode import ScanMode


@dataclass
class ScanContext:
    scan_mode: Union[ScanMode, str]
    command_path: str
    extra_headers: Optional[Dict[str, str]] = None
    target_path: Optional[Path] = None

    def __post_init__(self) -> None:
        self.command_id = str(uuid.uuid4())
        self.os_name, self.os_version = get_os_info()
        self.python_version = platform.python_version()

    def _get_repository_url(
        self,
    ) -> Optional[str]:
        repo_url = None
        if self.scan_mode == ScanMode.CI:
            repo_url = get_repository_url_from_ci()
        if repo_url is None and self.target_path is not None:
            repo_url = get_repository_url_from_path(self.target_path)
        return repo_url

    def get_http_headers(self) -> Dict[str, str]:
        """
        Returns the extra headers to send in HTTP requests.
        If `command_id` is not None, a `GGShield-Command-Id` header will be sent.
        Adds the "GGShield-" prefix to the header's names.
        """
        headers = {
            "Version": __version__,
            "Command-Path": self.command_path,
            "Command-Id": self.command_id,
            "OS-Name": self.os_name,
            "OS-Version": self.os_version,
            "Python-Version": self.python_version,
        }
        repo_url = self._get_repository_url()
        if repo_url is not None:
            headers["Repository-URL"] = repo_url
        if self.extra_headers:
            headers = {**headers, **self.extra_headers}

        return {
            **{f"GGShield-{key}": str(value) for key, value in headers.items()},
            "mode": (
                self.scan_mode.value
                if isinstance(self.scan_mode, ScanMode)
                else self.scan_mode
            ),
        }



import os
from typing import Any, Optional, Tuple

import click

from ggshield.cmd.hmsl.hmsl_common_options import (
    full_hashes_option,
    naming_strategy_option,
)
from ggshield.cmd.hmsl.hmsl_utils import check_secrets
from ggshield.cmd.utils.common_options import (
    add_common_options,
    json_option,
    text_json_format_option,
)
from ggshield.core import ui
from ggshield.core.errors import UnexpectedError
from ggshield.core.text_utils import pluralize
from ggshield.verticals.hmsl.collection import NamingStrategy, collect_list, prepare
from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.api_client import (
    VaultAPIClient,
)
from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.cli import (
    get_vault_cli_token,
)
from ggshield.verticals.hmsl.secret_manager.hashicorp_vault.exceptions import (
    VaultCliTokenFetchingError,
    VaultForbiddenItemError,
)


def _get_vault_token(use_vault_cli_token: bool) -> str:
    """
    Get the Vault token to use.
    """
    if use_vault_cli_token:
        try:
            return get_vault_cli_token()
        except VaultCliTokenFetchingError as exc:
            raise click.UsageError(
                f"could not get the token from Vault CLI: {str(exc)}"
            ) from exc

    env_token = os.getenv("VAULT_TOKEN")
    if env_token is None:
        raise click.UsageError(
            "you need to specify the Vault token to use, either through the VAULT_TOKEN"
            " environment variable or by using --use-cli-token to use the token "
            "from the Vault CLI."
        )

    return env_token


def _split_vault_mount_and_path(initial_path: str) -> Tuple[str, str]:
    """
    From a given initial path like secret/my_app/prod/env, split it in two:
    first the mount name, then the path.
    """
    split_path = initial_path.lstrip("/").split("/")
    return (split_path[0], "/".join(split_path[1:]).strip("/"))


@click.command()
@click.option(
    "--use-cli-token",
    "use_cli_token",
    is_flag=True,
    show_default=True,
    default=False,
    help="Instead of getting the token from the environment variable, "
    "get it from the CLI tool.",
)
@click.option(
    "--url",
    "url",
    required=False,
    type=str,
    help="The URL of the secret manager server.",
)
@click.option(
    "--recursive",
    "-r",
    "recursive",
    is_flag=True,
    show_default=True,
    default=False,
    help="If the secret manager path is a directory and not a file, explore recursively.",
)
@click.argument(
    "vault_path",
    type=str,
)
@add_common_options()
@text_json_format_option
@json_option
@full_hashes_option
@naming_strategy_option
@click.pass_context
def check_hashicorp_vault_cmd(
    ctx: click.Context,
    use_cli_token: bool,
    url: Optional[str],
    recursive: bool,
    vault_path: str,
    full_hashes: bool,
    naming_strategy: NamingStrategy,
    **kwargs: Any,
) -> int:
    """
    Check secrets of an Hashicorp Vault instance.
    Only compatible with the kv secret engines (v1 or v2) for now.

    Will use the VAULT_URL environment variable to get the Vault instance URL or
    the --url option if no environment variable is set.

    Will use the VAULT_TOKEN environment variable to authenticate, except
    if the --use-cli-token option is set.
    """

    # Get the Vault URL
    if url is None:
        url = os.getenv("VAULT_URL")
        if url is None:
            raise click.UsageError(
                "you need to specify the URL of your Vault, "
                "either through --url or in the VAULT_URL environment variable"
            )

    vault_token = _get_vault_token(use_cli_token)  # noqa: F841
    vault_client = VaultAPIClient(url, vault_token)

    # Get mount object and check it exists and we have access to it
    mount_name, secret_path = _split_vault_mount_and_path(vault_path)
    all_kv_mounts = list(vault_client.get_kv_mounts())
    mount = next((item for item in all_kv_mounts if item.name == mount_name), None)
    if mount is None:
        raise UnexpectedError(
            f"mount {mount_name} not found. Make sure it exists and "
            "that your token has access to it."
        )

    ui.display_info("Fetching secrets from Vault...")
    try:
        result = vault_client.get_secrets(mount, secret_path, recursive)
    # catch this error here if the path provided is a file that we don't have access to
    except VaultForbiddenItemError:
        raise UnexpectedError(
            f"access to the given path '{vault_path}' was forbidden. "
            "Are you sure the permissions of the token are correct?"
        )

    if len(result.not_fetched_paths) > 0:
        ui.display_error(
            f"Could not fetch {len(result.not_fetched_paths)} paths. "
            "Make sure your token has access to all the secrets in your vault."
        )
        if ui.is_verbose():
            ui.display_error("> The following paths could not be fetched:")
            for path in result.not_fetched_paths:
                ui.display_error(f"- {path}")
    ui.display_info(
        f"Got {len(result.secrets)} {pluralize('secret', len(result.secrets))}."
    )

    collected_secrets = collect_list(result.secrets)
    # full_hashes is True because we need the hashes to decrypt the secrets.
    # They will correctly be truncated by our client later (same as normal check cmd)
    prepared_secrets = prepare(collected_secrets, naming_strategy, full_hashes=True)

    check_secrets(
        ctx=ctx,
        prepared_secrets=prepared_secrets,
        full_hashes=full_hashes,
    )

    return 0



from pathlib import Path

import pytest
from click.testing import CliRunner

from ggshield.__main__ import cli
from ggshield.verticals.hmsl.crypto import hash_string
from tests.unit.conftest import assert_invoke_exited_with, assert_invoke_ok


RESULTS_CONTENT = (
    '{"hint": "f7f17c88638b42465b6c620a0c7648ef470e611c1fdf90166aac613601799f81", "payload": '
    '"KzzhJEq/pM2RaWav9NAvjw45Gfp/26UGDDzYiDNuOCup0PfoAHNViOGX14/a7uUNWLk53zGIar3s2xOW/xxzMEgTT2owjH52gGalhwKBfTY="}\n'
    '{"hint": "71d27eee3aa6c1751110ea338f23a5cfe11da717ea27453e7fe09e1594c3f8e7", "payload": '
    '"zmrGtuhTtgxNkk9SA250HTxXQ+mfoJQlZ76CPx50juK4XFCCTbFNv6ZeahGRQqW4+vf92DEwpGTHVzjiQEF6JebJsoRuaMQDSntHQ17z0UU="}\n'
    '{"hint": "f1b2fcaf134f3a08513ec6603ee3281511f349166fea5ef3356dd62051a76aa8", "payload": '
    '"Qzp7zRkFeIlPhiy6VMeUyo4vaCJAFmuqDQITH4WFC1BH51eHDNcL1UOw5u8dKmBWRJMfY7Zh7atyTl++hbsYDnIItJi8LFO5Yyzj+xte+ik="}\n'
    '{"hint": "89740ad4cd63fa9a7637325a7bef91c0ba93d0a45bbf687beb76bacaf5fa8da3", "payload": '
    '"kUlYx2lO5dOtFAM7XPT7uyk5v81ajJeg7Uepq1D4oyWQcf3ijMRThqsMrkKkUXSXHcAL182yCAgbub/NDF2wFA+Lyr5qBdb3qBBFLztfFz0="}\n'
    '{"hint": "3be2d605a3d143bfea373887ed16e7935be0e3c189cbee4d343c92ed6c89fdb8", "payload": '
    '"GZIG82jOLH5gXB5NNJt7NyfUOQUpk720wA3LItmVrXKCIK2PursytFkg/pPtzBXyPifNZtsOaNf5an+5Pz3mVysVMoCF9dXGFt1AFRi8lXk="}\n'
    '{"hint": "16787b637f7787685909539f65cc100b591d8c8d1074d0e5491aab33f364c86b", "payload": '
    '"4XgUM9pXWrLbQ8tH0AH7Za3u7tObAmlDXBSgwS+IE2m/NeDn3y7KF5H7yPB/faFDfKFirNiijhEfkBgfCz+FmZhDLCCzsga6hZN0S9he6EM="}\n'
    '{"hint": "e9ecc350e213860e9472057339443c830581c53e2b4dfb3aaa7e5fa4a854d5a3", "payload": '
    '"UDIP09t3tSk2IyQhxnJmF2gaDxhOY4zgrGpOzLeakIOZEmRxlyXYfdN3uFuTutnfdT7ZY+2Am2Q0Vst0L3EfuvomNdx/yL3desUApHq5o5I="}\n'
    '{"hint": "31ded0b51235ebde7d5fa10685d33b95e8a20a4e284220351812ca98ed20836b", "payload": '
    '"+FuUB48xvYQg1VTf1Jvyif14T8rLJETu3L0y2SJa7fJ+P7HDTDf/ESH8pLjJmadyNB3vl3t8KS3VH+lveCae53yVY66LncUCwuXVKd9s7G0="}\n'
    '{"hint": "19b9ba15c838c44d8965ac2300718fd8f9e2a038ff3ca7b3982fae50ec4afbfa", "payload": '
    '"YKk5NCIkiS5tmab2lXO1V2mpsPbRC+vAsz+TNHroEcpo8b0YhEjy6SCUXWkYMm2mBUFz3Kmvkqqd59Pdj4EXmvqrl1yNV2LlCCoJGD91SUY="}\n'
    '{"hint": "23ef947812513a59de504af2e291f9bbab287b941e0551f442e63f19f979679d", "payload": '
    '"0XmzWJNyq3gHbeqb5+T5xSjuwP1qFdrIbvsW4K5Spk+Yn2mfBs92Z3ipEngis2nZMNS+K99h/sh3+hvqTH5T5Z0p/YnCd2f+1E4suGEbVnA="}\n'
    '{"hint": "9c9e78a410131e548c733e08b1de9a3dcccbe5cda970cb6ad740655b7741e7b3", "payload": '
    '"WDmh3FQvY+i5DO+6bWeOkY5J78jHBHCsEFjl9u1PEpftDS5Htzcc/dQqrzFcYvBwU+RbPLag2z/w7PBW+m472D9R1OExamCWs6MjN65j3L0="}\n'
    # No location returned for the following secret
    '{"hint": "60d8cf27924c89ac016d59f20a63bb6449c4f058abf9eeed93d36c3f83e75681", "payload": '
    '"W9F0FQDILc/IRkpevr7mPLF5nxPaq9LM6Z8vcyQBS2Wm9j+D/y6jX8F0LjOUQTs="}\n'
)

RESULTS_CLEARTEXT_CONTENT = (
    '{"hash": "743d9fde380b7064cc6a8d3071184fc47905cf7440e5615cd46c7b6cbfb46d47", '
    '"count": 14, "url": "https://github.com/edly-io/devstack/commit/ccfc9c2d63c29'
    '17be60a9fd2a4c36ff3a8b9bb8c#diff-e45e45baeda1c1e73482975a664062aa56f20c03dd9d64a827aba57775bed0d3L158"}'
)


@pytest.fixture
def mapping_path(cli_fs_runner, tmp_path: Path):
    """Prepare a mapping file"""
    mapping_path = tmp_path / "mapping.txt"
    secrets = ["foo", "bar", "password", "1234"]
    mapping = {hash_string(secret): secret for secret in secrets}
    mapping_path.write_text(
        "\n".join(f"{key}:{value}" for key, value in mapping.items())
    )
    return mapping_path


@pytest.fixture
def mapping_path_no_location(cli_fs_runner, tmp_path: Path):
    """Prepare a mapping file"""
    mapping_path_no_location = tmp_path / "mapping_no_location.txt"
    secrets = ["apikey_with_no_location"]
    mapping_no_location = {hash_string(secret): secret for secret in secrets}
    mapping_path_no_location.write_text(
        "\n".join(f"{key}:{value}" for key, value in mapping_no_location.items())
    )
    return mapping_path_no_location


@pytest.fixture
def results_path(mapping_path: Path):
    """Prepare a results file"""
    results_path = mapping_path.parent / "results.txt"
    results_path.write_text(RESULTS_CONTENT)
    return results_path


@pytest.fixture
def full_hash_result(mapping_path: Path):
    """Prepare a results file"""
    results_path = mapping_path.parent / "results.txt"
    results_path.write_text(RESULTS_CLEARTEXT_CONTENT)
    return results_path


@pytest.mark.parametrize(
    "command",
    [
        ["hmsl", "decrypt"],
        ["hmsl", "decrypt", "none.txt"],
        ["hmsl", "decrypt", "-m", "none.txt"],
        ["hmsl", "decrypt", "-m", "none.txt", "void.txt"],
    ],
)
def test_hmsl_decrypt_no_files(cli_fs_runner: CliRunner, command) -> None:
    """
    GIVEN a cli
    WHEN running on non-existing files or other issues
    THEN the return code is 2
    """
    result = cli_fs_runner.invoke(cli, command)
    assert_invoke_exited_with(result, 2)


def test_hmsl_decrypt_default_behavior(
    cli_fs_runner: CliRunner, mapping_path, results_path: Path
) -> None:
    """
    GIVEN some secrets
    WHEN running the decrypt command on a file
    THEN the secrets are correctly decrypted
    """
    result = cli_fs_runner.invoke(
        cli, ["hmsl", "decrypt", "-m", str(mapping_path), str(results_path)]
    )
    assert_invoke_ok(result)
    assert result.output.count("> Secret ") == 1
    assert 'Secret name: "foo"' in result.output
    assert "URL: " in result.output


def test_hmsl_decrypt_full_hashes_behavior(
    cli_fs_runner: CliRunner, mapping_path, full_hash_result: Path
) -> None:
    """
    GIVEN a some full hashes response
    WHEN running the decrypt command on a file
    THEN the command accepts the decrypted payloads seamlessly
    """
    result = cli_fs_runner.invoke(
        cli, ["hmsl", "decrypt", "-m", str(mapping_path), str(full_hash_result)]
    )
    assert_invoke_ok(result)
    assert result.output.count("> Secret ") == 1
    assert 'Secret name: "password"' in result.output


def test_hmsl_decrypt_no_location_returned(
    cli_fs_runner: CliRunner, mapping_path_no_location, results_path: Path
) -> None:
    """
    GIVEN a secret for which no location is returned
    WHEN running the decrypt command on a file
    THEN the secret is correctly decrypted with no location associated
    """
    result = cli_fs_runner.invoke(
        cli, ["hmsl", "decrypt", "-m", str(mapping_path_no_location), str(results_path)]
    )
    assert result.output.count("> Secret ") == 1
    assert 'Secret name: "apikey_with_no_location"' in result.output
    assert "URL: " not in result.output



"""
High level log and debug configuration.

core.ui.log_utils contains code which is independent of ggshield itself.

This module contains ggshield-specific logging code, such as logging arguments and
version numbers, or manipulating the log level of some ggshield dependencies.
"""

import logging
import sys
from typing import Optional

import pygitguardian

import ggshield
from ggshield.core import ui
from ggshield.core.ui import log_utils


logger = logging.getLogger(__name__)


def setup_debug_mode(*, filename: Optional[str] = None) -> None:
    """
    Enable debug mode: set up logger and set the UI level to DEBUG.
    """
    ui.set_level(ui.Level.DEBUG)

    log_utils.set_log_handler(filename)

    # Silence charset_normalizer, its debug output does not bring much
    logging.getLogger("charset_normalizer").setLevel(logging.WARNING)

    logger.debug("args=%s", sys.argv)
    logger.debug("ggshield=%s", ggshield.__version__)
    logger.debug("py-gitguardian=%s", pygitguardian.__version__)



from pathlib import Path
from typing import Any, Dict, Optional

import click
from pygitguardian.models import Detail, HoneytokenResponse

from ggshield.cmd.honeytoken.utils import generate_random_honeytoken_name
from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.client import create_client_from_config
from ggshield.core.errors import UnexpectedError
from ggshield.utils.click import RealPath


def _dict_to_string(data: Dict[str, str], space: bool = False) -> str:
    """Returns a string with 'key=value' for each key-value pair in the dictionary."""
    space_char = " " if space else ""
    return "\n".join([f"{k}{space_char}={space_char}{v}" for k, v in data.items()])


@click.command()
@click.option(
    "--name",
    "name",
    required=False,
    type=str,
    help="Specify a name for your honeytoken. If this option is not provided, a unique name will be generated with a \
`ggshield-` prefix.",
)
@click.option(
    "--type",
    "type_",
    required=True,
    type=click.Choice(("AWS",)),
    help="Specify the type of honeytoken that you want to create. (For now only AWS honeytokens are supported)",
)
@click.option(
    "--description",
    "description",
    required=False,
    type=str,
    help="Add a description to your honeytoken (250 characters max).",
)
@click.option(
    "-o",
    "--output",
    "output_file",
    type=RealPath(file_okay=True, dir_okay=False, readable=True, writable=True),
    required=False,
    help="Specify a filename to append your honeytoken directly to the content of this file. \
If the file does not exist, it will be created.",
)
@add_common_options()
@click.pass_context
def create_cmd(
    ctx: click.Context,
    name: Optional[str],
    type_: str,
    description: Optional[str],
    output_file: Optional[Path],
    **kwargs: Any,
) -> int:
    """
    Command to create a honeytoken.

    The prerequisites to use this command are the following:

    - you have the necessary permissions as a user (for now, Honeytoken is restricted to
    users with a "Manager" access level),

    - the personal access token used by ggshield has the `honeytokens:write` scope.
    """
    # if name is not given, generate a random one
    if not name:
        name = generate_random_honeytoken_name()
    ctx_obj = ContextObj.get(ctx)
    client = create_client_from_config(ctx_obj.config)

    response = client.create_honeytoken(name, type_, description)
    if not isinstance(response, (Detail, HoneytokenResponse)):
        raise UnexpectedError("Unexpected honeytoken response")

    if (
        isinstance(response, Detail)
        and response.status_code == 403
        and "allowlist" not in response.detail
    ):
        raise UnexpectedError(
            """ggshield does not have permissions to create honeytokens on your workspace. Make sure that:

- the honeytoken module is enabled for your GitGuardian workspace,
- you have the necessary permissions as a user,
- the personal access token used by ggshield has the `honeytokens:write` scope.

To learn more, visit https://docs.gitguardian.com/honeytoken/getting-started."""
        )
    elif isinstance(response, Detail):
        raise UnexpectedError(response.detail)
    honeytoken: HoneytokenResponse = response

    token_to_display = {
        f"{honeytoken.type_}_{k}".lower(): v for k, v in honeytoken.token.items()
    }

    if output_file:
        click.echo(
            f"Your honeytoken has been created successfully in {output_file}\n"
            f'Your honeytoken #{honeytoken.id} "{honeytoken.name}" is accessible in your workspace: '
            f"{honeytoken.gitguardian_url}\n"
        )

        # special output for piped or redirect
        with output_file.open("a") as opened_output_file:
            opened_output_file.write(f"{_dict_to_string(token_to_display)}\n")

    else:
        click.echo(
            "Your honeytoken has been created successfully.\n"
            "You can now paste it into your code. We recommend that you place it in one unique place.\n"
            "-------------------------------\n"
            f"{_dict_to_string(token_to_display, space=True)}\n"
            "-------------------------------\n"
            f'Your honeytoken #{honeytoken.id} "{honeytoken.name}" is accessible in your workspace: '
            f"{honeytoken.gitguardian_url}\n"
        )

    return 0



from copy import deepcopy
from unittest import mock

import click
import pytest

from ggshield.core.config.user_config import SecretConfig
from ggshield.core.scan import StringScannable
from ggshield.utils.git_shell import Filemode
from ggshield.verticals.secret import Result, Results, SecretScanCollection
from ggshield.verticals.secret.output import SecretTextOutputHandler
from ggshield.verticals.secret.output.secret_text_output_handler import (
    format_line_count_break,
)
from ggshield.verticals.secret.secret_scan_collection import (
    IgnoreKind,
    IgnoreReason,
    group_secrets_by_ignore_sha,
)
from tests.factories import PolicyBreakFactory, ScannableFactory, ScanResultFactory
from tests.unit.conftest import (
    _MULTI_SECRET_ONE_LINE_PATCH,
    _MULTI_SECRET_ONE_LINE_PATCH_OVERLAY,
    _MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT,
    _MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT,
    _MULTI_SECRET_TWO_LINES_PATCH,
    _MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT,
    _ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
    _ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
    _SIMPLE_SECRET_MULTILINE_PATCH,
    _SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT,
    _SIMPLE_SECRET_PATCH,
    _SIMPLE_SECRET_PATCH_SCAN_RESULT,
)


@pytest.mark.parametrize(
    "show_secrets",
    [pytest.param(True, id="show_secrets"), pytest.param(False, id="hide_secrets")],
)
@pytest.mark.parametrize(
    "verbose",
    [pytest.param(True, id="verbose"), pytest.param(False, id="clip_long_lines")],
)
@pytest.mark.parametrize(
    "result_input",
    [
        pytest.param(
            Result.from_scan_result(
                StringScannable(
                    content=_SIMPLE_SECRET_PATCH,
                    url="leak.txt",
                    filemode=Filemode.NEW,
                ),
                scan_result=_SIMPLE_SECRET_PATCH_SCAN_RESULT,
                secret_config=SecretConfig(),
            ),
            id="_SIMPLE_SECRET_PATCH_SCAN_RESULT",
        ),
        pytest.param(
            Result.from_scan_result(
                StringScannable(
                    content=_MULTI_SECRET_ONE_LINE_PATCH,
                    url="leak.txt",
                    filemode=Filemode.NEW,
                ),
                scan_result=_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT,
                secret_config=SecretConfig(),
            ),
            id="_MULTI_SECRET_ONE_LINE_PATCH_SCAN_RESULT",
        ),
        pytest.param(
            Result.from_scan_result(
                StringScannable(
                    content=_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY,
                    url="leak.txt",
                    filemode=Filemode.NEW,
                ),
                scan_result=_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT,
                secret_config=SecretConfig(),
            ),
            id="_MULTI_SECRET_ONE_LINE_PATCH_OVERLAY_SCAN_RESULT",
        ),
        pytest.param(
            Result.from_scan_result(
                StringScannable(
                    content=_MULTI_SECRET_TWO_LINES_PATCH,
                    url="leak.txt",
                    filemode=Filemode.NEW,
                ),
                scan_result=_MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT,
                secret_config=SecretConfig(),
            ),
            id="_MULTI_SECRET_TWO_LINES_PATCH_SCAN_RESULT",
        ),
        pytest.param(
            Result.from_scan_result(
                StringScannable(
                    content=_SIMPLE_SECRET_MULTILINE_PATCH,
                    url="leak.txt",
                    filemode=Filemode.NEW,
                ),
                scan_result=_SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT,
                secret_config=SecretConfig(),
            ),
            id="_SIMPLE_SECRET_MULTILINE_PATCH_SCAN_RESULT",
        ),
        pytest.param(
            Result.from_scan_result(
                StringScannable(
                    content=_ONE_LINE_AND_MULTILINE_PATCH_CONTENT,
                    url="leak.txt",
                    filemode=Filemode.NEW,
                ),
                scan_result=_ONE_LINE_AND_MULTILINE_PATCH_SCAN_RESULT,
                secret_config=SecretConfig(),
            ),
            id="_ONE_LINE_AND_MULTILINE_PATCH_CONTENT",
        ),
    ],
)
def test_leak_message(result_input, snapshot, show_secrets, verbose):
    # The text output includes the version of the secrets engine, but this version is
    # None until we make an API call. Since this test does not make any API call, set
    # the version to a fake value.
    with mock.patch(
        "ggshield.verticals.secret.output.secret_text_output_handler.VERSIONS"
    ) as VERSIONS:
        VERSIONS.secrets_engine_version = "3.14.159"

        secret_config = SecretConfig(show_secrets=show_secrets)
        output_handler = SecretTextOutputHandler(
            secret_config=secret_config, verbose=verbose
        )

        # _process_scan_impl() modifies its SecretScanCollection arg(!), so make a copy of it
        new_result = deepcopy(result_input)

        output = output_handler._process_scan_impl(
            SecretScanCollection(
                id="scan",
                type="test",
                results=Results(results=[new_result], errors=[]),
                optional_header="> This is an example header",
            )
        )
    # remove colors because color codes are not the same on all OSes. This is required
    # to compare the output with snapshots.
    output = click.unstyle(output)

    snapshot.assert_match(output)

    # all ignore sha should be in the output
    assert all(
        ignore_sha in output
        for ignore_sha in group_secrets_by_ignore_sha(result_input.secrets)
    )


def assert_policies_displayed(output, verbose, ignore_known_secrets, secrets):
    for secret in secrets:
        if not ignore_known_secrets or verbose:
            # All secrets are displayed no matter if they're known or not
            assert f"Secret detected: {secret.detector}" in output
            if secret.known_secret:
                assert "Known by GitGuardian dashboard: YES" in output
                assert (
                    "https://dashboard.gitguardian.com/workspace/1/incidents/" in output
                )
            else:
                assert "Known by GitGuardian dashboard: NO" in output
                assert "Incident URL: N/A" in output
        else:
            if secret.known_secret:
                assert f"Secret detected: {secret.detector}" not in output

    if ignore_known_secrets:
        secrets_number = sum(1 for x in secrets if not x.known_secret)
    else:
        secrets_number = len(secrets)

    if secrets_number:
        assert (
            f"{secrets_number} incident{'s' if secrets_number > 1 else ''} detected"
            in output
        )


def assert_warning_is_displayed(
    output: str,
    ignore_known_secrets: bool,
    secrets_types: str,
    known_secrets_number: int,
):
    if ignore_known_secrets and secrets_types in {
        "only_known_secrets",
        "mixed_secrets",
    }:
        plural = (
            "s ignored because they are"
            if known_secrets_number > 1
            else " ignored because it is"
        )
        assert (
            f"Warning: {known_secrets_number} secret{plural} already known by "
            "your GitGuardian dashboard and you used the "
            "`--ignore-known-secrets` option." in output
        )


def assert_no_leak_message_is_diplayed(
    output: str, ignore_known_secrets: bool, secrets_types: str
):
    if secrets_types == "no_secrets":
        assert "No secrets have been found" in output
        assert "No new secrets have been found" not in output
    elif ignore_known_secrets and secrets_types == "only_known_secrets":
        assert "No new secrets have been found" in output
        assert "No secrets have been found" not in output
    else:
        assert "No secrets have been found" not in output
        assert "No new secrets have been found" not in output


def test_format_line_count_break():
    assert format_line_count_break(5) == "\x1b[36m\x1b[22m\x1b[22m  ...\n\x1b[0m"


@pytest.mark.parametrize(
    ("ignore_reason"),
    (
        None,
        IgnoreReason(kind=IgnoreKind.IGNORED_MATCH),
        IgnoreReason(kind=IgnoreKind.BACKEND_EXCLUDED, detail="some detail"),
    ),
)
def test_ignore_reason(ignore_reason):
    """
    GIVEN an result
    WHEN it is passed to the json output handler
    THEN the ignore_reason field is as expected
    """

    secret_config = SecretConfig()
    scannable = ScannableFactory()
    policy_break = PolicyBreakFactory(content=scannable.content)
    result = Result.from_scan_result(
        scannable, ScanResultFactory(policy_breaks=[policy_break]), secret_config
    )
    result.secrets[0].ignore_reason = ignore_reason

    output_handler = SecretTextOutputHandler(secret_config=secret_config, verbose=False)

    output = output_handler._process_scan_impl(
        SecretScanCollection(
            id="scan",
            type="scan",
            results=Results(results=[result], errors=[]),
        )
    )

    if ignore_reason is None:
        assert "Ignored:" not in output
    else:
        assert "Ignored:" in output
        assert ignore_reason.to_human_readable() in output



from typing import Any

import click

from ggshield.cmd.utils.common_options import add_common_options
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core.cache import Cache
from ggshield.core.config import Config
from ggshield.core.text_utils import pluralize
from ggshield.core.types import IgnoredMatch


@click.command()
@click.argument(
    "secret_sha",
    nargs=1,
    type=str,
    required=False,
    metavar="SECRET_SHA",
)
@click.option(
    "--last-found",
    is_flag=True,
    help="Ignore secrets found in the last `ggshield secret scan` run.",
)
@click.option(
    "--name",
    type=str,
    help="Name of the secret to ignore.",
    metavar="NAME",
)
@add_common_options()
@click.pass_context
def ignore_cmd(
    ctx: click.Context,
    secret_sha: str,
    name: str,
    last_found: bool,
    **kwargs: Any,
) -> None:
    """
    Ignore some secrets.

    The `secret ignore` command instructs ggshield to ignore secrets.

    Option `--name` allows to specify the name of the secret to ignore.

    Option `--last-found` ignores all secrets found during the last scan.

    The command adds the ignored secrets to the `secrets.ignored_matches` section of your local
    configuration file. If no local configuration file is found, a `.gitguardian.yaml` file is created.
    """

    ctx_obj = ContextObj.get(ctx)
    config = ctx_obj.config
    path = config.config_path

    if last_found:
        if secret_sha or name:
            raise click.UsageError(
                "Option `--last-found` cannot be used with `SECRET_SHA` or `--name`."
            )
        nb = ignore_last_found(config, ctx_obj.cache)
    else:
        if not name:
            raise click.UsageError(
                "Option `--name` is required when ignoring a secret."
            )
        ignored_match = IgnoredMatch(name=name, match=secret_sha)
        config.add_ignored_match(ignored_match)
        nb = 1

    config.save()
    secrets_word = pluralize("secret", nb)
    click.echo(
        f"Added {nb} {secrets_word} to the secret.ignored_matches section of {path}."
    )


def ignore_last_found(config: Config, cache: Cache) -> int:
    """
    Add last found secrets from .cache_ggshield into ignored_matches
    in the local .gitguardian.yaml config file so that they are ignored on next run
    Secrets are added as `hash`
    """
    for secret in cache.last_found_secrets:
        config.add_ignored_match(secret)
    return len(cache.last_found_secrets)



import json
import os
import platform
from pathlib import Path
from typing import Any, Dict

import pytest


# The directory holding ggshield repository checkout
ROOT_DIR = Path(__file__).parent.parent

JSON_SCHEMAS_DIR = ROOT_DIR / "doc/schemas"

# This is a test token, it is always reported as a valid secret
# Use your own value if needed
GG_VALID_TOKEN = os.getenv("TEST_GG_VALID_TOKEN", "ggtt-v-12345azert")  # ggignore
GG_VALID_TOKEN_IGNORE_SHA = os.getenv(
    "TEST_GG_VALID_TOKEN_IGNORE_SHA",
    "56c126cef75e3d17c3de32dac60bab688ecc384a054c2c85b688c1dd7ac4eefd",
)

# This secret must be a secret known by the dashboard running functional tests
KNOWN_SECRET = os.environ.get("TEST_KNOWN_SECRET", "")

# This secret must not be known by the dashboard running our tests
UNKNOWN_SECRET = os.environ.get("TEST_UNKNOWN_SECRET", "ggtt-v-0d4buhg879")  # ggignore


def is_windows():
    return platform.system() == "Windows"


skipwindows = pytest.mark.skipif(
    is_windows() and not os.environ.get("DISABLE_SKIPWINDOWS"),
    reason="Skipped on Windows for now, define DISABLE_SKIPWINDOWS environment variable to unskip",
)


@pytest.fixture(autouse=True)
def do_not_use_real_user_dirs(monkeypatch, tmp_path):
    """
    This fixture ensures we do not use real user directories.
    Overridden directories are:
    - the auth configuration directory, where `ggshield auth` stores credentials.
    - the cache directory
    - the home directory
    """
    monkeypatch.setenv("GG_CONFIG_DIR", str(tmp_path / "config"))
    monkeypatch.setenv("GG_CACHE_DIR", str(tmp_path / "cache"))
    monkeypatch.setenv("GG_DATA_DIR", str(tmp_path / "data"))
    monkeypatch.setenv("GG_USER_HOME_DIR", str(tmp_path / "home"))


@pytest.fixture(autouse=True)
def do_not_use_colors(monkeypatch):
    """
    This fixture ensures we do not print colors for easier testing.
    """
    monkeypatch.setenv("NO_COLOR", "1")


def clean_directory(path: Path):
    for filepath in path.iterdir():
        if filepath.is_file():
            os.remove(filepath)


@pytest.fixture(scope="session")
def secret_json_schema() -> Dict[str, Any]:
    """Load the JSON schema for all `secret scan` commands,"""
    return _load_json_schema("secret.json")


@pytest.fixture(scope="session")
def quota_json_schema() -> Dict[str, Any]:
    """Load the JSON schema for `quota` command."""
    return _load_json_schema("quota.json")


@pytest.fixture(scope="session")
def api_status_json_schema() -> Dict[str, Any]:
    """Load the JSON schema for `api-status` command."""
    return _load_json_schema("api-status.json")


@pytest.fixture(scope="session")
def config_list_json_schema() -> Dict[str, Any]:
    """Load the JSON schema for `config list` command."""
    return _load_json_schema("config_list.json")


def _load_json_schema(name: str) -> Dict[str, Any]:
    """Load a JSON schema and patch it to reject additional properties. We patch it this
    way to ensure all fields of ggshield JSON output are documented in the JSON schema.
    """
    with (JSON_SCHEMAS_DIR / name).open() as fp:
        dct = json.load(fp)
    _reject_additional_properties(dct)
    return dct


def _reject_additional_properties(dct: Dict[str, Any]):
    """Helper for JSON Schema fixtures: adds `"additionalProperties": false` to all
    objects of the JSON schema, ensuring we do not add fields without updating the
    schema.
    """
    try:
        type_ = dct["type"]
    except KeyError:
        return
    if type_ == "object":
        dct["additionalProperties"] = False
        for child in dct["properties"].values():
            _reject_additional_properties(child)
    elif type_ == "array":
        _reject_additional_properties(dct["items"])



"""
This module defines options which should be available on all commands, such as the
-v, --verbose option.

To use it:
- Add the `@add_common_options()` decorator after all the `click.option()` calls of the
  command function.
- Add a `**kwargs: Any` argument to the command function.

The `kwargs` argument is required because due to the way click works,
`add_common_options()` adds an argument for each option it defines.
"""

from typing import Any, Callable, List, Optional, TypeVar

import click

from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.debug import setup_debug_mode
from ggshield.cmd.utils.output_format import OutputFormat
from ggshield.core import ui
from ggshield.core.config.user_config import UserConfig


AnyFunction = Callable[..., Any]


# The argument of a Click option callback function
ArgT = TypeVar("ArgT")

# A Click option callback function
ClickCallback = Callable[
    [click.Context, click.Parameter, Optional[ArgT]], Optional[ArgT]
]


def get_config_from_context(ctx: click.Context) -> UserConfig:
    """Returns the UserConfig object stored in Click context"""
    return ContextObj.get(ctx).config.user_config


def create_ctx_callback(*option_names: str) -> ClickCallback[ArgT]:
    """Helper function to define a Click option callback for simple cases where we only
    have to set a value on Click context object if the option is defined.
    """

    def callback(
        ctx: click.Context, param: click.Parameter, value: Optional[ArgT]
    ) -> Optional[ArgT]:
        if value is not None and ctx.obj is not None:
            obj = ContextObj.get(ctx)
            for name in option_names[:-1]:
                obj = getattr(obj, name)
            setattr(obj, option_names[-1], value)
        return value

    return callback


def create_config_callback(*option_names: str) -> ClickCallback[ArgT]:
    """Helper function to define a Click option callback for simple cases where we only
    have to set a configuration attribute if the option is defined.

    to reach UserConfig.foo, set option_names to ["foo"]
    to reach Userconfig.secret.bar, set option_names to ["secret", "bar"]
    """

    def callback(
        ctx: click.Context, param: click.Parameter, value: Optional[ArgT]
    ) -> Optional[ArgT]:
        if value is not None and ctx.obj is not None:
            # If ctx.obj has not been defined yet, then it means we are in the top-level
            # cli() function and the config has not been loaded yet, so we can't change
            # it.
            #
            # cli() takes care of applying the few config-related options it receives
            # itself.
            obj = get_config_from_context(ctx)
            for name in option_names[:-1]:
                obj = getattr(obj, name)
            setattr(obj, option_names[-1], value)
        return value

    return callback


def verbose_callback(
    ctx: click.Context, param: click.Parameter, value: Optional[bool]
) -> Optional[bool]:
    if value is not None:
        ui.ensure_level(ui.Level.VERBOSE)
    return value


_verbose_option = click.option(
    "-v",
    "--verbose",
    is_flag=True,
    default=None,
    help="Verbose display mode.",
    callback=verbose_callback,
)


def debug_callback(
    ctx: click.Context, param: click.Parameter, value: Optional[bool]
) -> Optional[bool]:
    if value is not None:
        setup_debug_mode()
    return value


_debug_option = click.option(
    "--debug",
    is_flag=True,
    default=None,
    help="Send log output to stderr. Equivalent to `--log-file -`.",
    callback=debug_callback,
)


def log_file_callback(
    ctx: click.Context, param: click.Parameter, value: Optional[str]
) -> Optional[str]:
    if value is not None:
        setup_debug_mode(filename=value)
    return value


_log_file_option = click.option(
    "--log-file",
    metavar="FILE",
    help="Send log output to FILE. Use '-' to redirect to stderr.",
    envvar="GITGUARDIAN_LOG_FILE",
    callback=log_file_callback,
)


_allow_self_signed_option = click.option(
    "--allow-self-signed",
    is_flag=True,
    default=None,
    help="Ignore ssl verification.",
    callback=create_config_callback("allow_self_signed"),
)

_check_for_updates = click.option(
    "--check-for-updates/--no-check-for-updates",
    is_flag=True,
    default=None,
    help="After executing commands, check if a new version of ggshield is available.",
    callback=create_ctx_callback("check_for_updates"),
)


exit_zero_option = click.option(
    "--exit-zero",
    is_flag=True,
    default=None,
    envvar="GITGUARDIAN_EXIT_ZERO",
    help=(
        "Return a 0 (non-error) status code, even if incidents are found."
        " An error status code will still be returned for other errors, such as connection errors."
        " This option can also be set with the `GITGUARDIAN_EXIT_ZERO` environment"
        " variable."
    ),
    callback=create_config_callback("exit_zero"),
)


def add_common_options() -> Callable[[AnyFunction], AnyFunction]:
    def decorator(cmd: AnyFunction) -> AnyFunction:
        _verbose_option(cmd)
        _debug_option(cmd)
        _log_file_option(cmd)
        _allow_self_signed_option(cmd)
        _check_for_updates(cmd)
        return cmd

    return decorator


def _set_json_output_format(
    ctx: click.Context, param: click.Parameter, value: Optional[bool]
) -> Optional[bool]:
    if value:
        ctx_obj = ContextObj.get(ctx)
        ctx_obj.output_format = OutputFormat.JSON
    return value


json_option = click.option(
    "--json",
    "json_output",
    is_flag=True,
    default=None,
    help="Shorthand for `--format json`.",
    callback=_set_json_output_format,
)


def _set_output_format(
    ctx: click.Context, param: click.Parameter, value: Optional[str]
) -> Optional[str]:
    if value:
        ctx_obj = ContextObj.get(ctx)
        ctx_obj.output_format = OutputFormat(value)
    return value


def _create_format_option(
    formats: List[OutputFormat],
) -> Callable[[click.decorators.FC], click.decorators.FC]:
    return click.option(
        "--format",
        type=click.Choice([x.value for x in formats]),
        help="Format to use for the output.",
        callback=_set_output_format,
    )


# If a command only supports text and json formats, it should use this option
text_json_format_option = _create_format_option([OutputFormat.TEXT, OutputFormat.JSON])


# If a command supports text, sarif and json formats, it should use this option
text_json_sarif_format_option = _create_format_option(
    [OutputFormat.TEXT, OutputFormat.JSON, OutputFormat.SARIF]
)


instance_option = click.option(
    "--instance",
    required=False,
    type=str,
    help="URL of the instance to use.",
    metavar="URL",
    callback=create_ctx_callback("config", "cmdline_instance_name"),
)



from .default_command_group import DefaultCommandGroup
from .natural_order_group import NaturalOrderGroup
from .real_path import RealPath


__all__ = [
    "DefaultCommandGroup",
    "NaturalOrderGroup",
    "RealPath",
]



#!/usr/bin/python3
from typing import Any, Union

import click
from pygitguardian import GGClient
from pygitguardian.models import Detail, Quota, QuotaResponse

from ggshield.cmd.utils.common_options import (
    add_common_options,
    instance_option,
    json_option,
    text_json_format_option,
)
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.cmd.utils.quota import format_quota_color
from ggshield.core.client import create_client_from_config
from ggshield.core.errors import UnexpectedError


@click.command()
@text_json_format_option
@json_option
@instance_option
@add_common_options()
@click.pass_context
def quota_cmd(ctx: click.Context, **kwargs: Any) -> int:
    """
    Show the remaining quota of API calls available for the entire workspace.
    """
    ctx_obj = ContextObj.get(ctx)
    client: GGClient = create_client_from_config(ctx_obj.config)
    response: Union[Detail, QuotaResponse] = client.quota_overview()

    if not isinstance(response, (Detail, QuotaResponse)):
        raise UnexpectedError("Unexpected quota response")

    if isinstance(response, Detail):
        raise UnexpectedError(response.detail)

    quota: Quota = response.content

    click.echo(
        quota.to_json()
        if ctx_obj.use_json
        else (
            f"Quota available: {format_quota_color(quota.remaining, quota.limit)}\n"
            f"Quota used in the last 30 days: {quota.count}\n"
            f"Total Quota of the workspace: {quota.limit}\n"
        )
    )

    return 0



import subprocess
import sys
from pathlib import Path
from typing import Optional, Union

from ggshield.utils.git_shell import git


class Repository:
    """
    Helper class to create Git repositories, for test purposes
    """

    def __init__(self, path: Path, remote_url: Optional[str] = None):
        self.path = path
        self.remote_url = remote_url
        self._credentials_set = False

    def git(self, *args: Union[str, Path]) -> str:
        try:
            return git(["-C", str(self.path)] + [str(x) for x in args])
        except subprocess.CalledProcessError as exc:
            out = exc.stdout.decode("utf-8", errors="ignore")
            err = exc.stderr.decode("utf-8", errors="ignore")
            print(f"Command failed with return code {exc.returncode}", file=sys.stderr)
            print(f"\n# stdout\n\n{out}", file=sys.stderr)
            print(f"\n# stderr\n\n{err}", file=sys.stderr)
            raise exc

    @classmethod
    def create(cls, path: Path, bare=False, initial_branch="main") -> "Repository":
        cmd = ["init", str(path), "--initial-branch", initial_branch]
        if bare:
            cmd.append("--bare")
        git(cmd)
        return cls(path)

    @classmethod
    def clone(cls, url: Union[str, Path], path: Path) -> "Repository":
        git(["clone", str(url), str(path)])
        return cls(path, remote_url=str(url))

    def add(self, *args: Union[str, Path]):
        self.git("add", *args)

    def push(self, *args: str) -> None:
        self.git("push", *args)

    def create_commit(self, message: str = "Test commit") -> str:
        self._ensure_credentials_are_set()
        self.git("commit", "--allow-empty", "-m", message)
        return self.get_top_sha()

    def checkout(self, name: str) -> None:
        self.git("checkout", name)

    def create_branch(self, name: str, orphan: bool = False) -> None:
        self.git("checkout", "--orphan" if orphan else "-b", name)

    def get_top_sha(self) -> str:
        out = self.git("rev-parse", "HEAD")
        return out.strip()

    def _ensure_credentials_are_set(self):
        if self._credentials_set:
            return
        self.git("config", "user.name", "ggshield-test")
        self.git("config", "user.email", "ggshield-test@example.com")
        self._credentials_set = True

    def remove_unreachable_commits(self):
        # Remove all unreachable commits and their references.
        # It is used to simulate the CI env which cannot access detached history.
        self.git("reflog", "expire", "--expire-unreachable=now", "--all")
        self.git("gc", "--prune=now")


def create_pre_receive_repo(tmp_path) -> Repository:
    repo = Repository.create(tmp_path)
    repo.create_commit("initial commit")

    # Detach from the current branch to simulate what happens when pre-receive
    # is called: the new commits are not in any branch yet.
    repo.git("checkout", "--detach")
    return repo



import os
from pathlib import Path
from typing import Any

import click
from click import UsageError

from ggshield.cmd.secret.scan.secret_scan_common_options import (
    add_secret_scan_common_options,
    create_output_handler,
)
from ggshield.cmd.utils.common_decorators import exception_wrapper
from ggshield.cmd.utils.context_obj import ContextObj
from ggshield.core import ui
from ggshield.core.cache import ReadOnlyCache
from ggshield.core.client import create_client_from_config
from ggshield.core.git_hooks.ci import collect_commit_range_from_ci_env
from ggshield.core.scan import ScanContext, ScanMode
from ggshield.utils.git_shell import check_git_dir
from ggshield.verticals.secret.repo import scan_commit_range


@click.command()
@add_secret_scan_common_options()
@click.pass_context
@exception_wrapper
def ci_cmd(ctx: click.Context, **kwargs: Any) -> int:
    """
    Scan the set of pushed commits that triggered the CI pipeline.
    """
    ctx_obj = ContextObj.get(ctx)
    ctx_obj.client = create_client_from_config(ctx_obj.config)
    config = ctx_obj.config
    check_git_dir()
    if not (os.getenv("CI") or os.getenv("JENKINS_HOME") or os.getenv("BUILD_BUILDID")):
        raise UsageError("`secret scan ci` should only be used in a CI environment.")

    commit_list, ci_mode = collect_commit_range_from_ci_env()
    mode_header = f"{ScanMode.CI.value}/{ci_mode.value}"

    ui.display_verbose(f"Commits to scan: {len(commit_list)}")

    scan_context = ScanContext(
        scan_mode=mode_header,
        command_path=ctx.command_path,
        target_path=Path.cwd(),
        extra_headers={"Ci-Mode": ci_mode.name},
    )

    return scan_commit_range(
        client=ctx_obj.client,
        cache=ReadOnlyCache(),
        commit_list=commit_list,
        output_handler=create_output_handler(ctx),
        exclusion_regexes=ctx_obj.exclusion_regexes,
        secret_config=config.user_config.secret,
        scan_context=scan_context,
    )
